# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
# STATEMENT OF CHANGES: This file is derived from sources licensed under the FreeSurfer 1.0 license
# terms, and this file has been changed.
# The full licensing terms of the original work are found at:
# https://github.com/freesurfer/freesurfer/blob/2995ded957961a7f3704de57eee88eb6cc30d52d/LICENSE.txt
# A copy of the license has been archived in the ORIGINAL_LICENSE file
# found within this redistribution.
#
# The original file this work derives from is found at:
# https://github.com/freesurfer/freesurfer/blob/2995ded957961a7f3704de57eee88eb6cc30d52d/mri_synthstrip/mri_synthstrip
#
# [April 2022] CHANGES:
#    * MAINT: Split the monolithic file into model and CLI submodules
#    * ENH: Replace freesurfer Python bundle with in-house code.
#
"""
Robust, universal skull-stripping for brain images of any type.
If you use SynthStrip in your analysis, please cite:

  A Hoopes, JS Mora, AV Dalca, B Fischl, M Hoffmann.
  SynthStrip: Skull-Stripping for Any Brain Image.
  https://arxiv.org/abs/2203.09974

"""


def main():
    """Entry point to SynthStrip."""
    import os
    from argparse import ArgumentParser
    import numpy as np
    import scipy
    import nibabel as nb
    import torch
    from .model import StripModel

    # parse command line
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "-i",
        "--image",
        metavar="file",
        required=True,
        help="Input image to skullstrip.",
    )
    parser.add_argument(
        "-o", "--out", metavar="file", help="Save stripped image to path."
    )
    parser.add_argument(
        "-m", "--mask", metavar="file", help="Save binary brain mask to path."
    )
    parser.add_argument("-g", "--gpu", action="store_true", help="Use the GPU.")
    parser.add_argument(
        "-b",
        "--border",
        default=1,
        type=int,
        help="Mask border threshold in mm. Default is 1.",
    )
    parser.add_argument("--model", metavar="file", help="Alternative model weights.")
    args = parser.parse_args()

    # sanity check on the inputs
    if not args.out and not args.mask:
        parser.fatal("Must provide at least --out or --mask output flags.")

    # necessary for speed gains (I think)
    torch.backends.cudnn.benchmark = True
    torch.backends.cudnn.deterministic = True

    # configure GPU device
    if args.gpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        device = torch.device("cuda")
        device_name = "GPU"
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        device = torch.device("cpu")
        device_name = "CPU"

    # configure model
    print(f"Configuring model on the {device_name}")

    with torch.no_grad():
        model = StripModel()
        model.to(device)
        model.eval()

    # load model weights
    if args.model is not None:
        modelfile = args.model
        print("Using custom model weights")
    else:
        raise RuntimeError("A model must be provided.")

    checkpoint = torch.load(modelfile, map_location=device)
    model.load_state_dict(checkpoint["model_state_dict"])

    # load input volume
    print(f"Input image read from: {args.image}")

    # normalize intensities
    image = nb.load(args.image)
    conformed = conform(image)
    in_data = conformed.get_fdata(dtype="float32")
    in_data -= in_data.min()
    in_data = np.clip(in_data / np.percentile(in_data, 99), 0, 1)
    in_data = in_data[np.newaxis, np.newaxis]

    # predict the surface distance transform
    input_tensor = torch.from_numpy(in_data).to(device)
    with torch.no_grad():
        sdt = model(input_tensor).cpu().numpy().squeeze()

    # unconform the sdt and extract mask
    sdt_target = resample_like(
        nb.Nifti1Image(sdt, conformed.affine, None),
        image,
        output_dtype="int16",
        cval=100,
    )
    sdt_data = np.asanyarray(sdt_target.dataobj).astype("int16")

    # find largest CC (just do this to be safe for now)
    components = scipy.ndimage.label(sdt_data.squeeze() < args.border)[0]
    bincount = np.bincount(components.flatten())[1:]
    mask = components == (np.argmax(bincount) + 1)
    mask = scipy.ndimage.morphology.binary_fill_holes(mask)

    # write the masked output
    if args.out:
        img_data = image.get_fdata()
        bg = np.min([0, img_data.min()])
        img_data[mask == 0] = bg
        nb.Nifti1Image(img_data, image.affine, image.header).to_filename(
            args.out,
        )
        print(f"Masked image saved to: {args.out}")

    # write the brain mask
    if args.mask:
        hdr = image.header.copy()
        hdr.set_data_dtype("uint8")
        nb.Nifti1Image(mask, image.affine, hdr).to_filename(args.mask)
        print(f"Binary brain mask saved to: {args.mask}")

    print("If you use SynthStrip in your analysis, please cite:")
    print("----------------------------------------------------")
    print("SynthStrip: Skull-Stripping for Any Brain Image.")
    print("A Hoopes, JS Mora, AV Dalca, B Fischl, M Hoffmann.")


def conform(input_nii):
    """Resample image as SynthStrip likes it."""
    import numpy as np
    import nibabel as nb
    from nitransforms.linear import Affine

    shape = np.array(input_nii.shape[:3])
    affine = input_nii.affine

    # Get corner voxel centers in index coords
    corner_centers_ijk = (
        np.array(
            [
                (i, j, k)
                for k in (0, shape[2] - 1)
                for j in (0, shape[1] - 1)
                for i in (0, shape[0] - 1)
            ]
        )
        + 0.5
    )

    # Get corner voxel centers in mm
    corners_xyz = (
        affine
        @ np.hstack((corner_centers_ijk, np.ones((len(corner_centers_ijk), 1)))).T
    )

    # Target affine is 1mm voxels in LIA orientation
    target_affine = np.diag([-1.0, 1.0, -1.0, 1.0])[:, (0, 2, 1, 3)]

    # Target shape
    extent = corners_xyz.min(1)[:3], corners_xyz.max(1)[:3]
    target_shape = ((extent[1] - extent[0]) / 1.0 + 0.999).astype(int)

    # SynthStrip likes dimensions be multiple of 64 (192, 256, or 320)
    target_shape = np.clip(
        np.ceil(np.array(target_shape) / 64).astype(int) * 64, 192, 320
    )

    # Ensure shape ordering is LIA too
    target_shape[2], target_shape[1] = target_shape[1:3]

    # Coordinates of center voxel do not change
    input_c = affine @ np.hstack((0.5 * (shape - 1), 1.0))
    target_c = target_affine @ np.hstack((0.5 * (target_shape - 1), 1.0))

    # Rebase the origin of the new, plumb affine
    target_affine[:3, 3] -= target_c[:3] - input_c[:3]

    nii = Affine(
        reference=nb.Nifti1Image(np.zeros(target_shape), target_affine, None),
    ).apply(input_nii)
    return nii


def resample_like(image, target, output_dtype=None, cval=0):
    """Resample the input image to be in the target's grid via identity transform."""
    from nitransforms.linear import Affine

    return Affine(reference=target).apply(image, output_dtype=output_dtype, cval=cval)


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""mriqc nipype interfaces """

from mriqc.interfaces.anatomical import (
    ArtifactMask,
    ComputeQI2,
    Harmonize,
    RotationMask,
    StructuralQC,
)
from mriqc.interfaces.bids import IQMFileSink
from mriqc.interfaces.common import ConformImage, EnsureSize
from mriqc.interfaces.functional import FunctionalQC, Spikes
from mriqc.interfaces.viz import PlotContours, PlotMosaic, PlotSpikes
from mriqc.interfaces.webapi import UploadIQMs

__all__ = [
    "ArtifactMask",
    "ComputeQI2",
    "ConformImage",
    "EnsureSize",
    "FunctionalQC",
    "Harmonize",
    "IQMFileSink",
    "PlotContours",
    "PlotMosaic",
    "PlotSpikes",
    "RotationMask",
    "Spikes",
    "StructuralQC",
    "UploadIQMs",
]



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Encapsulates report generation functions."""
from mriqc import messages


def individual_html(in_iqms, in_plots=None, api_id=None):
    import datetime
    from json import load
    from pathlib import Path

    from .. import config
    from ..data import IndividualTemplate
    from ..reports import REPORT_TITLES
    from ..reports.utils import iqms2html, read_report_snippet
    from ..utils.misc import BIDS_COMP

    def _get_details(in_iqms, modality):
        in_prov = in_iqms.pop("provenance", {})
        warn_dict = in_prov.pop("warnings", None)
        sett_dict = in_prov.pop("settings", None)

        wf_details = []
        if modality == "bold":
            bold_exclude_index = in_iqms.get("dumb_trs")
            if bold_exclude_index is None:
                config.loggers.cli.warning(
                    "Building bold report: no exclude index was found"
                )
            elif bold_exclude_index > 0:
                msg = """\
<span class="problematic">Non-steady state (strong T1 contrast) has been detected in the \
first {} volumes</span>. They were excluded before generating any QC measures and plots."""
                wf_details.append(msg.format(bold_exclude_index))

            wf_details.append(
                "Framewise Displacement was computed using <code>3dvolreg</code> (AFNI)"
            )

            fd_thres = sett_dict.pop("fd_thres")
            if fd_thres is not None:
                wf_details.append(
                    "Framewise Displacement threshold was defined at %f mm" % fd_thres
                )
        elif modality in ("T1w", "T2w"):
            if warn_dict.pop("small_air_mask", False):
                wf_details.append(
                    '<span class="problematic">Detected hat mask was too small</span>'
                )

            if warn_dict.pop("large_rot_frame", False):
                wf_details.append(
                    '<span class="problematic">Detected a zero-filled frame, has the original '
                    "image been rotated?</span>"
                )

        return in_prov, wf_details, sett_dict

    in_iqms = Path(in_iqms)
    with in_iqms.open() as jsonfile:
        iqms_dict = load(jsonfile)

    # Now, the in_iqms file should be correctly named
    out_file = str(Path(in_iqms.with_suffix(".html").name).resolve())

    # Extract and prune metadata
    metadata = iqms_dict.pop("bids_meta", None)
    mod = metadata.pop("modality", None)
    prov, wf_details, _ = _get_details(iqms_dict, mod)

    file_id = [metadata.pop(k, None) for k in list(BIDS_COMP.keys())]
    file_id = [comp for comp in file_id if comp is not None]

    if in_plots is None:
        in_plots = []
    else:
        if any(("melodic_reportlet" in k for k in in_plots)):
            REPORT_TITLES["bold"].insert(3, ("ICA components", "ica-comps"))
        if any(("plot_spikes" in k for k in in_plots)):
            REPORT_TITLES["bold"].insert(3, ("Spikes", "spikes"))

        in_plots = [
            (REPORT_TITLES[mod][i] + (read_report_snippet(v),))
            for i, v in enumerate(in_plots)
        ]

    pred_qa = None  # metadata.pop('mriqc_pred', None)
    _config = {
        "modality": mod,
        "dataset": metadata.pop("dataset", None),
        "bids_name": in_iqms.with_suffix("").name,
        "timestamp": datetime.datetime.now().strftime("%Y-%m-%d, %H:%M"),
        "version": config.environment.version,
        "imparams": iqms2html(iqms_dict, "iqms-table"),
        "svg_files": in_plots,
        "workflow_details": wf_details,
        "webapi_url": prov.pop("webapi_url"),
        "webapi_port": prov.pop("webapi_port"),
        "provenance": iqms2html(prov, "provenance-table"),
        "md5sum": prov["md5sum"],
        "metadata": iqms2html(metadata, "metadata-table"),
        "pred_qa": pred_qa,
    }

    if _config["metadata"] is None:
        _config["workflow_details"].append(
            '<span class="warning">File has no metadata</span> '
            "<span>(sidecar JSON file missing or empty)</span>"
        )

    tpl = IndividualTemplate()
    tpl.generate_conf(_config, out_file)

    end_message = messages.INDIVIDUAL_REPORT_GENERATED.format(out_file=out_file)
    config.loggers.cli.info(end_message)
    return out_file



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Visualizing resource recordings."""
import numpy as np
from matplotlib import pyplot as plt
import pandas as pd

_TIME_LABEL = "runtime"


def plot(filename, param="mem_vsm_mb", mask_processes=tuple(), out_file=None):
    """Plot a recording file."""
    data = pd.read_csv(filename, sep=r"\s+", comment="#")

    # Rebase all events to be relative to start and convert to seconds
    data["timestamp"] -= data["timestamp"][0]
    data["timestamp"] /= 1.0e9

    # Convert processes from rows to columns
    unified = pd.DataFrame({_TIME_LABEL: sorted(set(data["timestamp"].values))})

    pids = sorted(set(data["pid"].values))
    proc_names = []
    for pid in pids:
        pid_info = data[data["pid"] == pid]
        try:
            label = f"{pid_info['name'].values[0]}"
        except KeyError:
            label = f"{pid}"

        if label in mask_processes:
            continue

        rows = unified[_TIME_LABEL].isin(pid_info["timestamp"])
        if label not in unified.columns:
            proc_names.append(label)
            unified[label] = 0

        unified.loc[rows, label] += (
            pid_info[param].values
            / (1024 if param.startswith("mem") else 1)
        )

    unified[proc_names] = unified[proc_names].replace({0: np.nan})

    fig = plt.figure(figsize=(15, 10), facecolor="white")
    ax = plt.gca()
    _ = unified.plot.area(x=_TIME_LABEL, cmap="tab20", linewidth=0, ax=ax)
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    ax.set_xlabel("Run time (mm:ss)")
    if param.startswith("mem"):
        ax.set_ylabel("Memory fingerprint (GB)")
    elif param.startswith("cpu"):
        ax.set_ylabel("CPU utilization")
    else:
        ax.set_ylabel(param)

    ax.set_xticklabels(
        [f"{int(float(v) // 60)}:{int(float(v) % 60)}" for v in ax.get_xticks()]
    )
    if out_file is not None:
        fig.savefig(out_file, bbox_inches="tight", pad_inches=0, dpi=300)
        return

    return fig



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Manipulate Python warnings."""
import logging
import warnings

_wlog = logging.getLogger("py.warnings")
_wlog.addHandler(logging.NullHandler())


def _warn(message, category=None, stacklevel=1, source=None):
    """Redefine the warning function."""
    if category is not None:
        category = type(category).__name__
        category = category.replace("type", "WARNING")

    logging.getLogger("py.warnings").warning(f"{category or 'WARNING'}: {message}")


def _showwarning(message, category, filename, lineno, file=None, line=None):
    _warn(message, category=category)


warnings.warn = _warn
warnings.showwarning = _showwarning

# warnings.filterwarnings("ignore", category=FutureWarning)
# warnings.filterwarnings("ignore", category=DeprecationWarning)
# warnings.filterwarnings("ignore", category=ResourceWarning)
# # cmp is not used by mriqc, so ignore nipype-generated warnings
# warnings.filterwarnings("ignore", "cmp not installed")
# warnings.filterwarnings(
#     "ignore", "This has not been fully tested. Please report any failures."
# )
# warnings.filterwarnings("ignore", "sklearn.externals.joblib is deprecated in 0.21")
# warnings.filterwarnings("ignore", "can't resolve package from __spec__ or __package__")



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Reports."""
import nibabel as nb
import numpy as np
from mriqc import config
from mriqc.reports.individual import individual_html
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    InputMultiObject,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)


class _IndividualReportInputSpec(BaseInterfaceInputSpec):
    in_iqms = File(exists=True)
    in_plots = InputMultiObject(File(exists=True))
    api_id = traits.Str()


class _IndividualReportOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class IndividualReport(SimpleInterface):
    """Builds a provenance dictionary."""

    input_spec = _IndividualReportInputSpec
    output_spec = _IndividualReportOutputSpec

    def _run_interface(self, runtime):
        self._results["out_file"] = individual_html(
            self.inputs.in_iqms,
            self.inputs.in_plots if isdefined(self.inputs.in_plots) else None,
            self.inputs.api_id if isdefined(self.inputs.api_id) else None,
        )
        return runtime


class _AddProvenanceInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, desc="input file")
    air_msk = File(exists=True, desc="air mask file")
    rot_msk = File(exists=True, desc="rotation mask file")
    modality = traits.Str(mandatory=True, desc="provenance type")


class _AddProvenanceOutputSpec(TraitedSpec):
    out_prov = traits.Dict()


class AddProvenance(SimpleInterface):
    """Builds a provenance dictionary."""

    input_spec = _AddProvenanceInputSpec
    output_spec = _AddProvenanceOutputSpec

    def _run_interface(self, runtime):
        from nipype.utils.filemanip import hash_infile

        self._results["out_prov"] = {
            "md5sum": hash_infile(self.inputs.in_file),
            "version": config.environment.version,
            "software": "mriqc",
            "webapi_url": config.execution.webapi_url,
            "webapi_port": config.execution.webapi_port,
            "settings": {
                "testing": config.execution.debug,
            },
        }

        if self.inputs.modality in ("T1w", "T2w"):
            air_msk_size = (
                np.asanyarray(nb.load(self.inputs.air_msk).dataobj).astype(bool).sum()
            )
            rot_msk_size = (
                np.asanyarray(nb.load(self.inputs.rot_msk).dataobj).astype(bool).sum()
            )
            self._results["out_prov"]["warnings"] = {
                "small_air_mask": bool(air_msk_size < 5e5),
                "large_rot_frame": bool(rot_msk_size > 500),
            }

        if self.inputs.modality == "bold":
            self._results["out_prov"]["settings"].update(
                {
                    "fd_thres": config.workflow.fd_thres,
                }
            )

        return runtime



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
r"""
Measures for the spatial information
====================================

Definitions are given in the
:ref:`summary of structural IQMs <iqms_t1w>`.

.. _iqms_efc:

- **Entropy-focus criterion** (:py:func:`~mriqc.qc.anatomical.efc`).

.. _iqms_fber:

- **Foreground-Background energy ratio** (:py:func:`~mriqc.qc.anatomical.fber`,  [Shehzad2015]_).

.. _iqms_fwhm:

- **Full-width half maximum smoothness** (``fwhm_*``, see [Friedman2008]_).

.. _iqms_snr:

- **Signal-to-noise ratio** (:py:func:`~mriqc.qc.anatomical.snr`).

.. _iqms_summary:

- **Summary statistics** (:py:func:`~mriqc.qc.anatomical.summary_stats`).


Measures for the temporal information
-------------------------------------

.. _iqms_dvars :

DVARS
  D referring to temporal derivative of timecourses, VARS referring to
  RMS variance over voxels ([Power2012]_ ``dvars_nstd``) indexes the rate of change of
  BOLD signal across the entire brain at each frame of data. DVARS is calculated
  `with nipype
  <http://nipype.readthedocs.io/en/latest/interfaces/generated/nipype.algorithms.confounds.html#computedvars>`_
  after motion correction:

  .. math ::

      \text{DVARS}_t = \sqrt{\frac{1}{N}\sum_i \left[x_{i,t} - x_{i,t-1}\right]^2}


  .. note ::

    Intensities are scaled to 1000 leading to the units being expressed in x10
    :math:`\%\Delta\text{BOLD}` change.

  .. note ::

    MRIQC calculates two additional standardized values of the DVARS.
    The ``dvars_std`` metric is normalized with the standard deviation of the
    temporal difference time series. The ``dvars_vstd`` is a voxel-wise
    standardization of DVARS, where the temporal difference time series is
    normalized across time by that voxel standard deviation across time, before
    computing the RMS of the temporal difference [Nichols2013]_.

.. _iqms_gcor:

Global Correlation (``gcor``)
  calculates an optimized summary of time-series
  correlation as in [Saad2013]_ using AFNI's ``@compute_gcor``:

  .. math ::

      \text{GCOR} = \frac{1}{N}\mathbf{g}_u^T\mathbf{g}_u

  where :math:`\mathbf{g}_u` is the average of all unit-variance time series in a
  :math:`T` (# timepoints) :math:`\times` :math:`N` (# voxels) matrix.

.. _iqms_tsnr:

Temporal SNR (:abbr:`tSNR (temporal SNR)`, ``tsnr``)
  is a simplified interpretation of the tSNR definition [Kruger2001]_.
  We report the median value
  of the `tSNR map
  <http://nipype.readthedocs.io/en/latest/interfaces/generated/nipype.algorithms.confounds.html#tsnr>`_
  calculated like:

  .. math ::

      \text{tSNR} = \frac{\langle S \rangle_t}{\sigma_t},

  where :math:`\langle S \rangle_t` is the average BOLD signal (across time),
  and :math:`\sigma_t` is the corresponding temporal standard-deviation map. Higher
  values are better.


Measures for artifacts and other
--------------------------------

.. _iqms_fd:

Framewise Displacement
  expresses instantaneous head-motion [Jenkinson2002]_.
  MRIQC reports the average FD, labeled as ``fd_mean``.
  Rotational displacements are calculated as the displacement on the surface of a
  sphere of radius 50 mm [Power2012]_:

  .. math ::

      \text{FD}_t = |\Delta d_{x,t}| + |\Delta d_{y,t}| +
      |\Delta d_{z,t}| + |\Delta \alpha_t| + |\Delta \beta_t| + |\Delta \gamma_t|

  Along with the base framewise displacement, MRIQC reports the
  **number of timepoints above FD threshold** (``fd_num``), and the
  **percent of FDs above the FD threshold** w.r.t. the full timeseries (``fd_perc``).
  In both cases, the threshold is set at 0.20mm.

.. _iqms_gsr:

Ghost to Signal Ratio (:py:func:`~mriqc.qc.functional.gsr`)
  labeled in the reports as ``gsr_x`` and ``gsr_y``
  (calculated along the two possible phase-encoding axes **x**, **y**):

  .. math ::

      \text{GSR} = \frac{\mu_G - \mu_{NG}}{\mu_S}

  .. image :: ../_static/epi-gsrmask.png
    :width: 200px
    :align: center

.. _iqms_aor:

AFNI's outlier ratio (``aor``)
  Mean fraction of outliers per fMRI volume
  as given by AFNI's ``3dToutcount``.

.. _iqms_aqi:

AFNI's quality index (``aqi``)
  Mean quality index as computed by AFNI's ``3dTqual``; for each volume,
  it is one minus the Spearman's (rank) correlation of that volume with the
  median volume. Lower values are better.

.. _iqms_dummy:

Number of *dummy* scans** (``dummy``)
  A number of volumes in the begining of the
  fMRI timeseries identified as non-steady state.

.. topic:: References

  .. [Atkinson1997] Atkinson et al., *Automatic correction of motion artifacts
    in magnetic resonance images using an entropy
    focus criterion*, IEEE Trans Med Imag 16(6):903-910, 1997.
    doi:`10.1109/42.650886 <http://dx.doi.org/10.1109/42.650886>`_.

  .. [Friedman2008] Friedman, L et al., *Test--retest and between‐site reliability in a multicenter
    fMRI study*. Hum Brain Mapp, 29(8):958--972, 2008. doi:`10.1002/hbm.20440
    <http://dx.doi.org/10.1002/hbm.20440>`_.

  .. [Giannelli2010] Giannelli et al., *Characterization of Nyquist ghost in
    EPI-fMRI acquisition sequences implemented on two clinical 1.5 T MR scanner
    systems: effect of readout bandwidth and echo spacing*. J App Clin Med Phy,
    11(4). 2010.
    doi:`10.1120/jacmp.v11i4.3237 <http://dx.doi.org/10.1120/jacmp.v11i4.3237>`_.

  .. [Jenkinson2002] Jenkinson et al., *Improved Optimisation for the Robust and
    Accurate Linear Registration and Motion Correction of Brain Images*.
    NeuroImage, 17(2), 825-841, 2002.
    doi:`10.1006/nimg.2002.1132 <http://dx.doi.org/10.1006/nimg.2002.1132>`_.

  .. [Kruger2001] Krüger et al., *Physiological noise in oxygenation-sensitive
    magnetic resonance imaging*, Magn. Reson. Med. 46(4):631-637, 2001.
    doi:`10.1002/mrm.1240 <http://dx.doi.org/10.1002/mrm.1240>`_.

  .. [Nichols2013] Nichols, `Notes on Creating a Standardized Version of DVARS
    <http://www2.warwick.ac.uk/fac/sci/statistics/staff/academic-research/nichols/scripts/fsl/standardizeddvars.pdf>`_,
    2013.

  .. [Power2012] Power et al., *Spurious but systematic correlations in
    functional connectivity MRI networks arise from subject motion*,
    NeuroImage 59(3):2142-2154,
    2012, doi:`10.1016/j.neuroimage.2011.10.018
    <http://dx.doi.org/10.1016/j.neuroimage.2011.10.018>`_.

  .. [Saad2013] Saad et al. *Correcting Brain-Wide Correlation Differences
    in Resting-State FMRI*, Brain Conn 3(4):339-352,
    2013, doi:`10.1089/brain.2013.0156
    <http://dx.doi.org/10.1089/brain.2013.0156>`_.
"""
import os.path as op

import numpy as np

RAS_AXIS_ORDER = {"x": 0, "y": 1, "z": 2}


def gsr(epi_data, mask, direction="y", ref_file=None, out_file=None):
    """
    Compute the :abbr:`GSR (ghost to signal ratio)` [Giannelli2010]_.

    The procedure is as follows:

      #. Create a Nyquist ghost mask by circle-shifting the original mask by :math:`N/2`.

      #. Rotate by :math:`N/2`

      #. Remove the intersection with the original mask

      #. Generate a non-ghost background

      #. Calculate the :abbr:`GSR (ghost to signal ratio)`


    .. warning ::

      This should be used with EPI images for which the phase
      encoding direction is known.

    :param str epi_file: path to epi file
    :param str mask_file: path to brain mask
    :param str direction: the direction of phase encoding (x, y, all)
    :return: the computed gsr

    """
    direction = direction.lower()
    if direction[-1] not in ["x", "y", "all"]:
        raise Exception(
            "Unknown direction {}, should be one of x, -x, y, -y, all".format(direction)
        )

    if direction == "all":
        result = []
        for newdir in ["x", "y"]:
            ofile = None
            if out_file is not None:
                fname, ext = op.splitext(ofile)
                if ext == ".gz":
                    fname, ext2 = op.splitext(fname)
                    ext = ext2 + ext
                ofile = "{0}_{1}{2}".format(fname, newdir, ext)
            result += [gsr(epi_data, mask, newdir, ref_file=ref_file, out_file=ofile)]
        return result

    # Roll data of mask through the appropriate axis
    axis = RAS_AXIS_ORDER[direction]
    n2_mask = np.roll(mask, mask.shape[axis] // 2, axis=axis)

    # Step 3: remove from n2_mask pixels inside the brain
    n2_mask = n2_mask * (1 - mask)

    # Step 4: non-ghost background region is labeled as 2
    n2_mask = n2_mask + 2 * (1 - n2_mask - mask)

    # Step 5: signal is the entire foreground image
    ghost = np.mean(epi_data[n2_mask == 1]) - np.mean(epi_data[n2_mask == 2])
    signal = np.median(epi_data[n2_mask == 0])
    return float(ghost / signal)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
from mriqc.bin import messages


def get_parser():
    """
    Build parser object.
    """
    from argparse import ArgumentParser, RawTextHelpFormatter

    parser = ArgumentParser(
        description="MRIQCWebAPI: Check entries.", formatter_class=RawTextHelpFormatter
    )
    parser.add_argument(
        "modality",
        action="store",
        choices=["T1w", "bold"],
        help="number of expected items in the database",
    )
    parser.add_argument(
        "expected",
        action="store",
        type=int,
        help="number of expected items in the database",
    )
    parser.add_argument(
        "--webapi-url",
        action="store",
        default="https://mriqc.nimh.nih.gov/api/v1/T1w",
        type=str,
        help="IP address where the MRIQC WebAPI is listening",
    )
    return parser


def main():
    """Entry point."""
    import logging

    from requests import get

    # Run parser
    MRIQC_LOG = logging.getLogger(__name__)
    opts = get_parser().parse_args()
    get_log_message = messages.WEBAPI_GET.format(address=opts.webapi_url)
    MRIQC_LOG.info(get_log_message)
    response = get(opts.webapi_url).json()
    n_records = response["_meta"]["total"]
    response_log_message = messages.WEBAPI_REPORT.format(n_records=n_records)
    MRIQC_LOG.info(response_log_message)
    assert opts.expected == n_records


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Utilities: Jinja2 templates."""
from io import open  # pylint: disable=W0622

from pkg_resources import resource_filename as pkgrf


class Template(object):
    """
    Utility class for generating a config file from a jinja template.
    https://github.com/oesteban/endofday/blob/f2e79c625d648ef45b08cc1f11fd0bd84342d604/endofday/core/template.py
    """

    def __init__(self, template_str):
        import jinja2

        self.template_str = template_str
        self.env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchpath="/"),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def compile(self, configs):
        """Generates a string with the replacements"""
        template = self.env.get_template(self.template_str)
        return template.render(configs)

    def generate_conf(self, configs, path):
        """Saves the oucome after replacement on the template to file"""
        output = self.compile(configs)
        with open(path, "w+") as output_file:
            output_file.write(output)


class IndividualTemplate(Template):
    """Specific template for the individual report"""

    def __init__(self):
        super(IndividualTemplate, self).__init__(
            pkgrf("mriqc", "data/reports/individual.html")
        )


class GroupTemplate(Template):
    """Specific template for the individual report"""

    def __init__(self):
        super(GroupTemplate, self).__init__(pkgrf("mriqc", "data/reports/group.html"))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Helper functions."""

import json
from collections import OrderedDict
from pathlib import Path

import pandas as pd

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

IMTYPES = {
    "T1w": "anat",
    "T2w": "anat",
    "bold": "func",
}

BIDS_COMP = OrderedDict(
    [
        ("subject_id", "sub"),
        ("session_id", "ses"),
        ("task_id", "task"),
        ("acq_id", "acq"),
        ("rec_id", "rec"),
        ("run_id", "run"),
    ]
)

BIDS_EXPR = """\
^sub-(?P<subject_id>[a-zA-Z0-9]+)(_ses-(?P<session_id>[a-zA-Z0-9]+))?\
(_task-(?P<task_id>[a-zA-Z0-9]+))?(_acq-(?P<acq_id>[a-zA-Z0-9]+))?\
(_rec-(?P<rec_id>[a-zA-Z0-9]+))?(_run-(?P<run_id>[a-zA-Z0-9]+))?\
"""


def reorder_csv(csv_file, out_file=None):
    """
    Put subject, session and scan in front of csv file

    :param str csv_file: the input csv file
    :param str out_file: if provided, a new csv file is created

    :return: the path to the file with the columns reordered


    """
    if isinstance(csv_file, list):
        csv_file = csv_file[-1]

    if out_file is None:
        out_file = csv_file

    dataframe = pd.read_csv(csv_file)
    cols = dataframe.columns.tolist()  # pylint: disable=no-member
    try:
        cols.remove("Unnamed: 0")
    except ValueError:
        # The column does not exist
        pass

    for col in ["scan", "session", "subject"]:
        cols.remove(col)
        cols.insert(0, col)

    dataframe[cols].to_csv(out_file)
    return out_file


def rotate_files(fname):
    """A function to rotate file names"""
    import glob
    import os
    import os.path as op

    name, ext = op.splitext(fname)
    if ext == ".gz":
        name, ext2 = op.splitext(fname)
        ext = ext2 + ext

    if not op.isfile(fname):
        return

    prev = glob.glob("{}.*{}".format(name, ext))
    prev.insert(0, fname)
    prev.append("{0}.{1:d}{2}".format(name, len(prev) - 1, ext))
    for i in reversed(list(range(1, len(prev)))):
        os.rename(prev[i - 1], prev[i])


def bids_path(subid, sesid=None, runid=None, prefix=None, out_path=None, ext="json"):
    import os.path as op

    fname = "{}".format(subid)
    if prefix is not None:
        if not prefix.endswith("_"):
            prefix += "_"
        fname = prefix + fname
    if sesid is not None:
        fname += "_ses-{}".format(sesid)
    if runid is not None:
        fname += "_run-{}".format(runid)

    if out_path is not None:
        fname = op.join(out_path, fname)
    return op.abspath(fname + "." + ext)


def generate_pred(derivatives_dir, output_dir, mod):
    """
    Reads the metadata in the JIQM (json iqm) files and
    generates a corresponding prediction CSV table
    """

    if mod != "T1w":
        return None

    # If some were found, generate the CSV file and group report
    jsonfiles = list(output_dir.glob("sub-*/**/%s/sub-*_%s.json" % (IMTYPES[mod], mod)))
    if not jsonfiles:
        return None

    headers = list(BIDS_COMP.keys()) + ["mriqc_pred"]
    predictions = {k: [] for k in headers}

    for jsonfile in jsonfiles:
        with open(jsonfile, "r") as jsondata:
            data = json.load(jsondata).pop("bids_meta", None)

        if data is None:
            continue

        for k in headers:
            predictions[k].append(data.pop(k, None))

    dataframe = pd.DataFrame(predictions).sort_values(by=list(BIDS_COMP.keys()))

    # Drop empty columns
    dataframe.dropna(axis="columns", how="all", inplace=True)

    bdits_cols = list(set(BIDS_COMP.keys()) & set(dataframe.columns.ravel()))

    # Drop duplicates
    dataframe.drop_duplicates(bdits_cols, keep="last", inplace=True)

    out_csv = Path(output_dir) / ("%s_predicted_qa_csv" % mod)
    dataframe[bdits_cols + ["mriqc_pred"]].to_csv(str(out_csv), index=False)
    return out_csv


def generate_tsv(output_dir, mod):
    """
    Generates a tsv file from all json files in the derivatives directory
    """

    # If some were found, generate the CSV file and group report
    out_tsv = output_dir / ("group_%s.tsv" % mod)
    jsonfiles = list(output_dir.glob("sub-*/**/%s/sub-*_%s.json" % (IMTYPES[mod], mod)))
    if not jsonfiles:
        return None, out_tsv

    datalist = []
    for jsonfile in jsonfiles:
        dfentry = _read_and_save(jsonfile)

        if dfentry is not None:
            bids_name = str(Path(jsonfile.name).stem)
            dfentry.pop("bids_meta", None)
            dfentry.pop("provenance", None)
            dfentry["bids_name"] = bids_name
            datalist.append(dfentry)

    dataframe = pd.DataFrame(datalist)
    cols = dataframe.columns.tolist()  # pylint: disable=no-member
    dataframe = dataframe.sort_values(by=["bids_name"])

    # Drop duplicates
    dataframe.drop_duplicates(["bids_name"], keep="last", inplace=True)

    # Set filename at front
    cols.insert(0, cols.pop(cols.index("bids_name")))
    dataframe[cols].to_csv(str(out_tsv), index=False, sep="\t")
    return dataframe, out_tsv


def _read_and_save(in_file):
    data = json.loads(Path(in_file).read_text())
    return data if data else None


def _flatten(in_dict, parent_key="", sep="_"):
    items = []
    for k, val in list(in_dict.items()):
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(val, MutableMapping):
            items.extend(list(_flatten(val, new_key, sep=sep).items()))
        else:
            items.append((new_key, val))
    return dict(items)


def _flatten_dict(indict):
    out_qc = {}
    for k, value in list(indict.items()):
        if not isinstance(value, dict):
            out_qc[k] = value
        else:
            for subk, subval in list(value.items()):
                if not isinstance(subval, dict):
                    out_qc["_".join([k, subk])] = subval
                else:
                    for ssubk, ssubval in list(subval.items()):
                        out_qc["_".join([k, subk, ssubk])] = ssubval
    return out_qc



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Parser."""
import re

from mriqc import config


def _build_parser():
    """Build parser object."""
    import sys
    import warnings
    from argparse import Action, ArgumentDefaultsHelpFormatter, ArgumentParser
    from functools import partial
    from pathlib import Path
    from shutil import which

    from packaging.version import Version

    from .version import check_latest, is_flagged

    class DeprecateAction(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            warnings.warn(f"Argument {option_string} is deprecated and is *ignored*.")
            delattr(namespace, self.dest)

    def _path_exists(path, parser):
        """Ensure a given path exists."""
        if path is None or not Path(path).exists():
            raise parser.error(f"Path does not exist: <{path}>.")
        return Path(path).expanduser().absolute()

    def _min_one(value, parser):
        """Ensure an argument is not lower than 1."""
        value = int(value)
        if value < 1:
            raise parser.error("Argument can't be less than one.")
        return value

    def _to_gb(value):
        scale = {"G": 1, "T": 10**3, "M": 1e-3, "K": 1e-6, "B": 1e-9}
        digits = "".join([c for c in value if c.isdigit()])
        n_digits = len(digits)
        units = value[n_digits:] or "G"
        return int(digits) * scale[units[0]]

    def _drop_sub(value):
        """
        Drop ``sub-`` prefix of participant labels.

        >>> _drop_sub("s060")
        's060'
        >>> _drop_sub("sub-s060")
        's060'

        """
        return re.sub(r"^sub-", "", str(value))

    def _bids_filter(value):
        from json import loads

        if value and Path(value).exists():
            return loads(Path(value).read_text())

    verstr = f"MRIQC v{config.environment.version}"
    currentv = Version(config.environment.version)

    parser = ArgumentParser(
        description=f"""\
MRIQC {config.environment.version}
Automated Quality Control and visual reports for Quality Assesment of structural \
(T1w, T2w) and functional MRI of the brain.

{config.DSA_MESSAGE}""",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    PathExists = partial(_path_exists, parser=parser)
    PositiveInt = partial(_min_one, parser=parser)

    # Arguments as specified by BIDS-Apps
    # required, positional arguments
    # IMPORTANT: they must go directly with the parser object
    parser.add_argument(
        "bids_dir",
        action="store",
        type=PathExists,
        help="The root folder of a BIDS valid dataset (sub-XXXXX folders should "
        "be found at the top level in this folder).",
    )
    parser.add_argument(
        "output_dir",
        action="store",
        type=Path,
        help="The directory where the output files "
        "should be stored. If you are running group level analysis "
        "this folder should be prepopulated with the results of the "
        "participant level analysis.",
    )
    parser.add_argument(
        "analysis_level",
        action="store",
        nargs="+",
        help="Level of the analysis that will be performed. "
        "Multiple participant level analyses can be run independently "
        "(in parallel) using the same output_dir.",
        choices=["participant", "group"],
    )

    # optional arguments
    parser.add_argument("--version", action="version", version=verstr)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose_count",
        action="count",
        default=0,
        help="Increases log verbosity for each occurrence, debug level is -vvv.",
    )

    # TODO: add 'mouse', 'macaque', and other populations once the pipeline is working
    parser.add_argument(
        "--species",
        action="store",
        type=str,
        default="human",
        choices=["human", "rat"],
        help="Use appropriate template for population",
    )

    g_bids = parser.add_argument_group("Options for filtering BIDS queries")
    g_bids.add_argument(
        "--participant-label",
        "--participant_label",
        action="store",
        nargs="+",
        type=_drop_sub,
        help="A space delimited list of participant identifiers or a single "
        "identifier (the sub- prefix can be removed).",
    )
    g_bids.add_argument(
        "--session-id",
        action="store",
        nargs="*",
        type=str,
        help="Filter input dataset by session ID.",
    )
    g_bids.add_argument(
        "--run-id",
        action="store",
        type=int,
        nargs="*",
        help="Filter input dataset by run ID (only integer run IDs are valid).",
    )
    g_bids.add_argument(
        "--task-id",
        action="store",
        nargs="*",
        type=str,
        help="Filter input dataset by task ID.",
    )
    g_bids.add_argument(
        "-m",
        "--modalities",
        action="store",
        nargs="*",
        help="Filter input dataset by MRI type.",
    )
    g_bids.add_argument("--dsname", type=str, help="A dataset name.")
    g_bids.add_argument(
        "--bids-database-dir",
        metavar="PATH",
        type=PathExists,
        help="Path to an existing PyBIDS database folder, for faster indexing "
        "(especially useful for large datasets).",
    )

    # General performance
    g_perfm = parser.add_argument_group("Options to handle performance")
    g_perfm.add_argument(
        "--nprocs",
        "--n_procs",
        "--n_cpus",
        "-n-cpus",
        action="store",
        type=PositiveInt,
        help="""\
Maximum number of simultaneously running parallel processes executed by *MRIQC* \
(e.g., several instances of ANTs' registration). \
However, when ``--nprocs`` is greater or equal to the ``--omp-nthreads`` option, \
it also sets the maximum number of threads that simultaneously running processes \
may aggregate (meaning, with ``--nprocs 16 --omp-nthreads 8`` a maximum of two \
8-CPU-threaded processes will be running at a given time). \
Under this mode of operation, ``--nprocs`` sets the maximum number of processors \
that can be assigned work within an *MRIQC* job, which includes all the processors \
used by currently running single- and multi-threaded processes. \
If ``None``, the number of CPUs available will be automatically assigned (which may \
not be what you want in, e.g., shared systems like a HPC cluster.""",
    )
    g_perfm.add_argument(
        "--omp-nthreads",
        "--ants-nthreads",
        action="store",
        type=PositiveInt,
        help="""\
Maximum number of threads that multi-threaded processes executed by *MRIQC* \
(e.g., ANTs' registration) can use. \
If ``None``, the number of CPUs available will be automatically assigned (which may \
not be what you want in, e.g., shared systems like a HPC cluster.""",
    )
    g_perfm.add_argument(
        "--mem",
        "--mem_gb",
        "--mem-gb",
        dest="memory_gb",
        action="store",
        type=_to_gb,
        help="Upper bound memory limit for MRIQC processes.",
    )
    g_perfm.add_argument(
        "--testing",
        dest="debug",
        action="store_true",
        default=False,
        help="Use testing settings for a minimal footprint.",
    )
    g_perfm.add_argument(
        "-f",
        "--float32",
        action="store_true",
        default=True,
        help="Cast the input data to float32 if it's represented in higher precision "
        "(saves space and improves perfomance).",
    )
    g_perfm.add_argument(
        "--pdb",
        dest="pdb",
        action="store_true",
        default=False,
        help="Open Python debugger (pdb) on exceptions.",
    )

    # Control instruments
    g_outputs = parser.add_argument_group("Instrumental options")
    g_outputs.add_argument(
        "-w",
        "--work-dir",
        action="store",
        type=Path,
        default=Path("work").absolute(),
        help="Path where intermediate results should be stored.",
    )
    g_outputs.add_argument("--verbose-reports", default=False, action="store_true")
    g_outputs.add_argument(
        "--write-graph",
        action="store_true",
        default=False,
        help="Write workflow graph.",
    )
    g_outputs.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not run the workflow.",
    )
    g_outputs.add_argument(
        "--resource-monitor",
        "--profile",
        dest="resource_monitor",
        action="store_true",
        default=False,
        help="Hook up the resource profiler callback to nipype.",
    )
    g_outputs.add_argument(
        "--use-plugin",
        action="store",
        default=None,
        type=Path,
        help="Nipype plugin configuration file.",
    )
    g_outputs.add_argument(
        "--no-sub",
        default=False,
        action="store_true",
        help="Turn off submission of anonymized quality metrics "
        "to MRIQC's metrics repository.",
    )
    g_outputs.add_argument(
        "--email",
        action="store",
        default="",
        type=str,
        help="Email address to include with quality metric submission.",
    )

    g_outputs.add_argument(
        "--webapi-url",
        action="store",
        type=str,
        help="IP address where the MRIQC WebAPI is listening.",
    )
    g_outputs.add_argument(
        "--webapi-port",
        action="store",
        type=int,
        help="Port where the MRIQC WebAPI is listening.",
    )

    g_outputs.add_argument(
        "--upload-strict",
        action="store_true",
        default=False,
        help="Upload will fail if upload is strict.",
    )

    # ANTs options
    g_ants = parser.add_argument_group("Specific settings for ANTs")
    g_ants.add_argument(
        "--ants-float",
        action="store_true",
        default=False,
        help="Use float number precision on ANTs computations.",
    )
    g_ants.add_argument(
        "--ants-settings",
        action="store",
        help="Path to JSON file with settings for ANTs.",
    )

    # Functional workflow settings
    g_func = parser.add_argument_group("Functional MRI workflow configuration")
    if which("melodic") is not None:
        g_func.add_argument(
            "--ica",
            action="store_true",
            default=False,
            help="Run ICA on the raw data and include the components "
            "in the individual reports (slow but potentially very insightful).",
        )
    g_func.add_argument(
        "--fft-spikes-detector",
        action="store_true",
        default=False,
        help="Turn on FFT based spike detector (slow).",
    )
    g_func.add_argument(
        "--fd_thres",
        action="store",
        default=0.2,
        type=float,
        help="Threshold on framewise displacement estimates to detect outliers.",
    )
    g_func.add_argument(
        "--deoblique",
        action="store_true",
        default=False,
        help="Deoblique the functional scans during head motion correction "
        "preprocessing.",
    )
    g_func.add_argument(
        "--despike",
        action="store_true",
        default=False,
        help="Despike the functional scans during head motion correction "
        "preprocessing.",
    )
    g_func.add_argument(
        "--start-idx",
        action=DeprecateAction,
        type=int,
        help="DEPRECATED Initial volume in functional timeseries that should be "
        "considered for preprocessing.",
    )
    g_func.add_argument(
        "--stop-idx",
        action=DeprecateAction,
        type=int,
        help="DEPRECATED Final volume in functional timeseries that should be "
        "considered for preprocessing.",
    )

    latest = check_latest()
    if latest is not None and currentv < latest:
        print(
            f"""\
You are using MRIQC v{currentv}, and a newer version is available: {latest}.""",
            file=sys.stderr,
        )

    _blist = is_flagged()
    if _blist[0]:
        _reason = _blist[1] or "unknown"
        print(
            f"""\
WARNING: This version of MRIQC ({config.environment.version}) has been FLAGGED
(reason: {_reason}).
That means some severe flaw was found in it and we strongly \
discourage its usage.""",
            file=sys.stderr,
        )

    return parser


def parse_args(args=None, namespace=None):
    """Parse args and run further checks on the command line."""
    from logging import DEBUG
    from contextlib import suppress

    from ..utils.bids import collect_bids_data

    parser = _build_parser()
    opts = parser.parse_args(args, namespace)
    config.execution.log_level = int(max(25 - 5 * opts.verbose_count, DEBUG))
    config.from_dict(vars(opts))

    # Load base plugin_settings from file if --use-plugin
    if opts.use_plugin is not None:
        from yaml import load as loadyml

        with open(opts.use_plugin) as f:
            plugin_settings = loadyml(f)
        _plugin = plugin_settings.get("plugin")
        if _plugin:
            config.nipype.plugin = _plugin
            config.nipype.plugin_args = plugin_settings.get("plugin_args", {})
            config.nipype.nprocs = config.nipype.plugin_args.get(
                "nprocs", config.nipype.nprocs
            )

    bids_dir = config.execution.bids_dir
    output_dir = config.execution.output_dir
    work_dir = config.execution.work_dir
    version = config.environment.version

    # Ensure input and output folders are not the same
    if output_dir == bids_dir:
        parser.error(
            "The selected output folder is the same as the input BIDS folder. "
            "Please modify the output path (suggestion: %s)."
            % bids_dir
            / "derivatives"
            / ("mriqc-%s" % version.split("+")[0])
        )

    if bids_dir in work_dir.parents:
        parser.error(
            "The selected working directory is a subdirectory of the input BIDS folder. "
            "Please modify the output path."
        )

    # Validate inputs
    # if not opts.skip_bids_validation:
    #     from ..utils.bids import validate_input_dir

    #     build_log.info(
    #         "Making sure the input data is BIDS compliant (warnings can be ignored in most "
    #         "cases)."
    #     )
    #     validate_input_dir(
    #         config.environment.exec_env, opts.bids_dir, opts.participant_label
    #     )

    # Setup directories
    config.execution.log_dir = output_dir / "logs"
    # Check and create output and working directories
    config.execution.log_dir.mkdir(exist_ok=True, parents=True)
    output_dir.mkdir(exist_ok=True, parents=True)
    work_dir.mkdir(exist_ok=True, parents=True)

    # Force initialization of the BIDSLayout
    config.execution.init()
    all_subjects = config.execution.layout.get_subjects()
    if config.execution.participant_label is None:
        config.execution.participant_label = all_subjects

    participant_label = set(config.execution.participant_label)
    missing_subjects = participant_label - set(all_subjects)
    if missing_subjects:
        parser.error(
            "One or more participant labels were not found in the BIDS directory: "
            f"{', '.join(missing_subjects)}."
        )

    config.execution.participant_label = sorted(participant_label)

    # Handle analysis_level
    analysis_level = set(config.workflow.analysis_level)
    if not config.execution.participant_label:
        analysis_level.add("group")
    config.workflow.analysis_level = list(analysis_level)

    # List of files to be run
    bids_filters = {
        "participant_label": config.execution.participant_label,
        "session": config.execution.session_id,
        "run": config.execution.run_id,
        "task": config.execution.task_id,
        "bids_type": config.execution.modalities,
    }
    config.workflow.inputs = {
        mod: files
        for mod, files in collect_bids_data(
            config.execution.layout, **bids_filters
        ).items()
        if files
    }

    # Check the query is not empty
    if not list(config.workflow.inputs.values()):
        _j = "\n *"
        parser.error(
            f"""\
Querying BIDS dataset at <{config.execution.bids_dir}> got an empty result.
Please, check out your currently set filters:
{_j.join([''] + [': '.join((k, str(v))) for k, v in bids_filters.items()])}"""
        )

    # Check no DWI or others are sneaked into MRIQC
    unknown_mods = set(config.workflow.inputs.keys()) - set(("T1w", "T2w", "bold"))
    if unknown_mods:
        parser.error(
            "MRIQC is unable to process the following modalities: "
            f'{", ".join(unknown_mods)}.'
        )

    # Estimate the biggest file size / leave 1GB if some file does not exist (datalad)
    with suppress(FileNotFoundError):
        config.workflow.biggest_file_gb = _get_biggest_file_size_gb(
            [i for sublist in config.workflow.inputs.values() for i in sublist]
        )

    # set specifics for alternative populations
    if opts.species.lower() != "human":
        config.workflow.species = opts.species
        # TODO: add other species once rats are working
        if opts.species.lower() == "rat":
            config.workflow.template_id = "Fischer344"
            # mean distance from the lateral edge to the center of the brain is
            # ~ PA:10 mm, LR:7.5 mm, and IS:5 mm (see DOI: 10.1089/089771503770802853)
            # roll movement is most likely to occur, so set to 7.5 mm
            config.workflow.fd_radius = 7.5
            config.workflow.headmask = "NoBET"
            # block uploads for the moment; can be reversed before wider release
            config.execution.no_sub = True


def _get_biggest_file_size_gb(files):
    import os

    max_size = 0
    for file in files:
        size = os.path.getsize(file) / (1024**3)
        if size > max_size:
            max_size = size
    return max_size



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""A :obj:`~nipype.interfaces.utility.base.IdentityInterface` with a grafted Datalad getter."""

from pathlib import Path
from nipype.interfaces.io import add_traits
from nipype.interfaces.base import (
    BaseInterface,
    Directory,
    DynamicTraitedSpec,
    File,
    isdefined,
)
from mriqc import config


class DataladIdentityInterface(BaseInterface):
    """Sneaks a ``datalad get`` in paths, if datalad is available."""

    _always_run = True

    input_spec = DynamicTraitedSpec

    def __init__(self, fields=None, dataset_path=None, mandatory_inputs=True, **inputs):
        super().__init__(**inputs)

        if not fields:
            raise ValueError("fields must be a non-empty list")

        add_traits(
            self.inputs,
            ["dataset_path"],
            trait_type=Directory(exists=True, mandatory=True),
        )

        # Each input must be in the fields.
        if set(inputs.keys()) - set(fields):
            raise ValueError(
                f"inputs not in the fields: {', '.join(set(inputs.keys()) - set(fields))}."
            )
        add_traits(
            self.inputs,
            fields,
            trait_type=File(nohash=True, mandatory=mandatory_inputs)
        )

        if dataset_path:
            inputs["dataset_path"] = dataset_path

        self.inputs.trait_set(**inputs)

    def _run_interface(self, runtime):
        inputs = self.inputs.get()
        dataset_path = inputs.pop("dataset_path")
        if (
            not isdefined(dataset_path)
            or not (Path(dataset_path) / ".datalad").exists()
        ):
            config.loggers.interface.info("Datalad interface without dataset path defined.")
            return runtime

        _dl_found = False
        try:
            from datalad.api import get

            _dl_found = True
        except ImportError:
            def get(*args, **kwargs):
                """Mock datalad get."""

        dataset_path = Path(dataset_path)
        for field, value in inputs.items():
            if not isdefined(value):
                continue

            _pth = Path(value)
            if not _pth.is_absolute():
                _pth = dataset_path / _pth

            _datalad_candidate = _pth.is_symlink() and not _pth.exists()
            if not _dl_found and _datalad_candidate:
                config.loggers.interface.warning("datalad was required but not found")
                return runtime

            if _datalad_candidate:
                try:
                    result = get(
                        _pth,
                        dataset=dataset_path
                    )
                except Exception as exc:
                    config.loggers.interface.warning(f"datalad get on {_pth} failed.")
                    if (
                        config.environment.exec_env == "docker"
                        and ("This repository is not initialized for use by git-annex, "
                             "but .git/annex/objects/ exists") in f"{exc}"
                    ):
                        config.loggers.interface.warning(
                            "Execution seems containerirzed with Docker, please make sure "
                            "you are not running as root. To do so, please add the argument "
                            "``-u $(id -u):$(id -g)`` to your command line."
                        )
                    else:
                        config.loggers.interface.warning(str(exc))
                else:
                    if result[0]["status"] == "error":
                        config.loggers.interface.warning(f"datalad get failed: {result}")

        return runtime

    def _outputs(self):
        return self.inputs

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        return self.inputs

    def _list_outputs(self):
        raise NotImplementedError



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
from urllib.parse import urlparse

from mriqc import config, messages
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    Bunch,
    File,
    SimpleInterface,
    Str,
    TraitedSpec,
    isdefined,
    traits,
)

SECRET_KEY = "8sSYVI0XjFqacEMZ8wF4"

# metadata whitelist
META_WHITELIST = [
    "AccelNumReferenceLines",
    "AccelerationFactorPE",
    "AcquisitionMatrix",
    "CogAtlasID",
    "CogPOID",
    "CoilCombinationMethod",
    "ContrastBolusIngredient",
    "ConversionSoftware",
    "ConversionSoftwareVersion",
    "DelayTime",
    "DeviceSerialNumber",
    "EchoTime",
    "EchoTrainLength",
    "EffectiveEchoSpacing",
    "FlipAngle",
    "GradientSetType",
    "HardcopyDeviceSoftwareVersion",
    "ImageType",
    "ImagingFrequency",
    "InPlanePhaseEncodingDirection",
    "InstitutionAddress",
    "InstitutionName",
    "Instructions",
    "InversionTime",
    "MRAcquisitionType",
    "MRTransmitCoilSequence",
    "MagneticFieldStrength",
    "Manufacturer",
    "ManufacturersModelName",
    "MatrixCoilMode",
    "MultibandAccelerationFactor",
    "NumberOfAverages",
    "NumberOfPhaseEncodingSteps",
    "NumberOfVolumesDiscardedByScanner",
    "NumberOfVolumesDiscardedByUser",
    "NumberShots",
    "ParallelAcquisitionTechnique",
    "ParallelReductionFactorInPlane",
    "PartialFourier",
    "PartialFourierDirection",
    "PatientPosition",
    "PercentPhaseFieldOfView",
    "PercentSampling",
    "PhaseEncodingDirection",
    "PixelBandwidth",
    "ProtocolName",
    "PulseSequenceDetails",
    "PulseSequenceType",
    "ReceiveCoilName",
    "RepetitionTime",
    "ScanOptions",
    "ScanningSequence",
    "SequenceName",
    "SequenceVariant",
    "SliceEncodingDirection",
    "SoftwareVersions",
    "TaskDescription",
    "TaskName",
    "TotalReadoutTime",
    "TotalScanTimeSec",
    "TransmitCoilName",
    "VariableFlipAngleFlag",
    "acq_id",
    "modality",
    "run_id",
    "subject_id",
    "task_id",
    "session_id",
]

PROV_WHITELIST = ["version", "md5sum", "software", "settings"]

HASH_BIDS = ["subject_id", "session_id"]


class UploadIQMsInputSpec(BaseInterfaceInputSpec):
    in_iqms = File(exists=True, mandatory=True, desc="the input IQMs-JSON file")
    url = Str(mandatory=True, desc="URL (protocol and name) listening")
    port = traits.Int(desc="MRIQCWebAPI service port")
    path = Str(desc="MRIQCWebAPI endpoint root path")
    email = Str(desc="set sender email")
    strict = traits.Bool(
        False, usedefault=True, desc="crash if upload was not succesfull"
    )


class UploadIQMsOutputSpec(TraitedSpec):
    api_id = traits.Either(
        None, traits.Str, desc="Id for report returned by the web api"
    )


class UploadIQMs(SimpleInterface):
    """
    Upload features to MRIQCWebAPI
    """

    input_spec = UploadIQMsInputSpec
    output_spec = UploadIQMsOutputSpec
    always_run = True

    def _run_interface(self, runtime):
        email = None
        if isdefined(self.inputs.email):
            email = self.inputs.email

        rawurl = self.inputs.url
        if "://" not in rawurl:
            rawurl = "http://"
        url = urlparse(rawurl)

        if not url.scheme.startswith("http"):
            raise RuntimeError('Tried an unknown protocol "%s"' % url.scheme)

        port = url.port
        if isdefined(self.inputs.port):
            port = self.inputs.port

        path = url.path
        if isdefined(self.inputs.path):
            path = self.inputs.path

        self._results["api_id"] = None

        response = upload_qc_metrics(
            self.inputs.in_iqms,
            url.netloc,
            path=path,
            scheme=url.scheme,
            port=port,
            email=email,
        )

        try:
            self._results["api_id"] = response.json()["_id"]
        except (AttributeError, KeyError, ValueError):
            # response did not give us an ID
            errmsg = (
                "QC metrics upload failed to create an ID for the record "
                "uplOADED. rEsponse from server follows: {}".format(response.text)
            )
            config.loggers.interface.warning(errmsg)

        if response.status_code == 201:
            config.loggers.interface.info(messages.QC_UPLOAD_COMPLETE)
            return runtime

        errmsg = "QC metrics failed to upload. Status %d: %s" % (
            response.status_code,
            response.text,
        )
        config.loggers.interface.warning(errmsg)
        if self.inputs.strict:
            raise RuntimeError(response.text)

        return runtime


def upload_qc_metrics(in_iqms, loc, path="", scheme="http", port=None, email=None):
    """
    Upload qc metrics to remote repository.

    :param str in_iqms: Path to the qc metric json file as a string
    :param str scheme: the protocol (either http or https)
    :param str email: email address to be included with the metric submission
    :param bool upload_strict: the client should fail if it's strict mode

    :return: either the response object if a response was successfully sent
             or it returns the string "No Response"
    :rtype: object


    """
    from copy import deepcopy
    from json import dumps, loads
    from pathlib import Path

    import requests

    if port is None:
        port = 443 if scheme == "https" else 80

    in_data = loads(Path(in_iqms).read_text())

    # Extract metadata and provenance
    meta = in_data.pop("bids_meta")

    # For compatibility with WebAPI. Shold be rolled back to int
    if meta.get("run_id", None) is not None:
        meta["run_id"] = "%d" % meta.get("run_id")

    prov = in_data.pop("provenance")

    # At this point, data should contain only IQMs
    data = deepcopy(in_data)

    # Check modality
    modality = meta.get("modality", "None")
    if modality not in ("T1w", "bold", "T2w"):
        errmsg = (
            'Submitting to MRIQCWebAPI: image modality should be "bold", "T1w", or "T2w", '
            '(found "%s")' % modality
        )
        return Bunch(status_code=1, text=errmsg)

    # Filter metadata values that aren't in whitelist
    data["bids_meta"] = {k: meta[k] for k in META_WHITELIST if k in meta}
    # Filter provenance values that aren't in whitelist
    data["provenance"] = {k: prov[k] for k in PROV_WHITELIST if k in prov}

    # Hash fields that may contain personal information
    data["bids_meta"] = _hashfields(data["bids_meta"])

    if email:
        data["provenance"]["email"] = email

    if path and not path.endswith("/"):
        path += "/"
        if path.startswith("/"):
            path = path[1:]

    headers = {"Authorization": SECRET_KEY, "Content-Type": "application/json"}

    webapi_url = f"{scheme}://{loc}:{port}/{path}{modality}"
    start_message = messages.QC_UPLOAD_START.format(url=webapi_url)
    config.loggers.interface.info(start_message)
    try:
        # if the modality is bold, call "bold" endpoint
        response = requests.post(webapi_url, headers=headers, data=dumps(data))
    except requests.ConnectionError as err:
        errmsg = (
            "QC metrics failed to upload due to connection error shown below:\n%s" % err
        )
        return Bunch(status_code=1, text=errmsg)

    return response


def _hashfields(data):
    from hashlib import sha256

    for name in HASH_BIDS:
        if name in data:
            data[name] = sha256(data[name].encode()).hexdigest()

    return data



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Instrumentation to profile resource utilization."""
from time import time_ns, sleep
from datetime import datetime
from tempfile import mkstemp
from pathlib import Path
from multiprocessing import Process, Event
from contextlib import suppress
import signal
import psutil

_MB = 1024.0**2
SAMPLE_ATTRS = (
    "pid",
    "name",
    # "cmdline",
    "cpu_num",
    "cpu_percent",
    "memory_info",
    "num_threads",
    "num_fds",
)


def sample(
    pid=None,
    recursive=True,
    attrs=SAMPLE_ATTRS,
    exclude=tuple(),
):
    """
    Probe process tree and snapshot current resource utilization.

    Parameters
    ----------
    pid : :obj:`int` or :obj:`None`
        The process ID that must be sampled. If ``None`` then it samples the
        current process from which ``sample()`` has been called.
    recursive : :obj:`bool`
        Whether the sampler should descend and explore the whole process tree.
    attrs : :obj:`iterable` of :obj:`str`
        A list of :obj:`psutil.Process` attribute names that will be retrieved when
        sampling.

    """
    proc_list = [psutil.Process(pid)]
    if proc_list and recursive:
        with suppress(psutil.NoSuchProcess):
            proc_list += proc_list[0].children(recursive=True)

    proc_info = []
    for process in proc_list:
        if process.pid in exclude:
            continue
        with suppress(psutil.NoSuchProcess):
            proc_info.append(process.as_dict(attrs=attrs))

    return proc_info


def parse_sample(datapoint, timestamp=None, attrs=SAMPLE_ATTRS):
    """Convert a sample dictionary into a list of string values."""
    retval = [f"{timestamp or time_ns()}"]

    for attr in attrs:
        value = datapoint.get(attr, None)
        if value is None:
            continue

        if attr == "cmdline":
            value = " ".join(value).replace("'", "\\'").replace('"', '\\"')
            value = [f"'{value}'"]
        elif attr == "memory_info":
            value = [f"{value.rss / _MB}", f"{value.vms / _MB}"]
        else:
            value = [f"{value}"]

        retval += value

    return retval


def sample2file(
    pid=None, recursive=True, timestamp=None, fd=None, flush=True, exclude=tuple()
):
    if fd is None:
        return

    print(
        "\n".join(
            [
                "\t".join(parse_sample(s, timestamp=timestamp))
                for s in sample(pid=pid, recursive=recursive, exclude=exclude)
            ]
        ),
        file=fd,
    )
    if flush:
        fd.flush()


class ResourceRecorder(Process):
    """Attach a ``Thread`` to sample a specific PID with a certain frequence."""

    def __init__(
        self, pid, frequency=0.2, log_file=None, exclude_probe=True, **process_kwargs
    ):
        Process.__init__(self, name="nipype_resmon", daemon=True, **process_kwargs)

        self._pid = pid
        """The process to be sampled."""
        self._logfile = log_file
        """An open file descriptor where results are dumped."""
        self._exclude = exclude_probe or tuple()
        """A list/tuple containing PIDs that should not be monitored."""
        self._freq_ns = int(max(frequency, 0.02) * 1e9)
        """Sampling frequency (stored in ns)."""
        self._done = Event()
        """Flag indicating if the process is marked to finish."""

        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def run(self, *args, **kwargs):
        """Core monitoring function, called by start()"""

        # Open file now, because it cannot be pickled.
        _logfile = (
            Path(self._logfile)
            if self._logfile
            else Path(mkstemp(prefix="prof-", suffix=".tsv")[1])
        )
        _logfile.parent.mkdir(parents=True, exist_ok=True)
        self._logfile = str(_logfile.absolute())

        _logfile = _logfile.open("w")

        # Write headers (comment trace + header row)
        _header = [
            f"# MRIQC Resource recorder started tracking PID {self._pid} "
            f"{datetime.now().strftime('(%Y/%m/%d; %H:%M:%S)')}",
            "\t".join(("timestamp", *SAMPLE_ATTRS)).replace(
                "memory_info", "mem_rss_mb\tmem_vsm_mb"
            ),
        ]
        print("\n".join(_header), file=_logfile)

        # Add self to exclude list if pertinent
        if self._exclude is True:
            self._exclude = (psutil.Process().pid,)

        # Initiate periodic sampling
        start_time = time_ns()
        wait_til = start_time
        while not self._done.is_set():
            try:
                sample2file(self._pid, fd=_logfile, timestamp=wait_til)
            except psutil.NoSuchProcess:
                print(
                    f"# MRIQC Resource recorder killed "
                    f"{datetime.now().strftime('(%Y/%m/%d; %H:%M:%S)')}",
                    file=_logfile,
                )
                _logfile.flush()
                _logfile.close()
                break

            wait_til += self._freq_ns
            sleep(max(0, (wait_til - time_ns()) / 1.0e9))

        _logfile.close()

    def stop(self, *args):
        # Tear-down process
        self._done.set()
        with Path(self._logfile).open("a") as f:
            f.write(
                f"# MRIQC Resource recorder finished "
                f"{datetime.now().strftime('(%Y/%m/%d; %H:%M:%S)')}",
            )



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
# STATEMENT OF CHANGES: This file is derived from sources licensed under the FreeSurfer 1.0 license
# terms, and this file has been changed.
# The full licensing terms of the original work are found at:
# https://github.com/freesurfer/freesurfer/blob/2995ded957961a7f3704de57eee88eb6cc30d52d/LICENSE.txt
# A copy of the license has been archived in the ORIGINAL_LICENSE file
# found within this redistribution.
#
# The original file this work derives from is found at:
# https://github.com/freesurfer/freesurfer/blob/2995ded957961a7f3704de57eee88eb6cc30d52d/mri_synthstrip/mri_synthstrip
#
# [April 2022] CHANGES:
#    * MAINT: Split the monolithic file into model and CLI submodules
#    * ENH: Replace freesurfer Python bundle with in-house code.
#
"""
Robust, universal skull-stripping for brain images of any type.
If you use SynthStrip in your analysis, please cite:

  A Hoopes, JS Mora, AV Dalca, B Fischl, M Hoffmann.
  SynthStrip: Skull-Stripping for Any Brain Image.
  https://arxiv.org/abs/2203.09974

"""

import torch
import torch.nn as nn
import numpy as np


class StripModel(nn.Module):
    def __init__(
        self,
        nb_features=16,
        nb_levels=7,
        feat_mult=2,
        max_features=64,
        nb_conv_per_level=2,
        max_pool=2,
        return_mask=False,
    ):

        super().__init__()

        # dimensionality
        ndims = 3

        # build feature list automatically
        if isinstance(nb_features, int):
            if nb_levels is None:
                raise ValueError(
                    "must provide unet nb_levels if nb_features is an integer"
                )
            feats = np.round(nb_features * feat_mult ** np.arange(nb_levels)).astype(
                int
            )
            feats = np.clip(feats, 1, max_features)
            nb_features = [
                np.repeat(feats[:-1], nb_conv_per_level),
                np.repeat(np.flip(feats), nb_conv_per_level),
            ]
        elif nb_levels is not None:
            raise ValueError("cannot use nb_levels if nb_features is not an integer")

        # extract any surplus (full resolution) decoder convolutions
        enc_nf, dec_nf = nb_features
        nb_dec_convs = len(enc_nf)
        final_convs = dec_nf[nb_dec_convs:]
        dec_nf = dec_nf[:nb_dec_convs]
        self.nb_levels = int(nb_dec_convs / nb_conv_per_level) + 1

        if isinstance(max_pool, int):
            max_pool = [max_pool] * self.nb_levels

        # cache downsampling / upsampling operations
        MaxPooling = getattr(nn, "MaxPool%dd" % ndims)
        self.pooling = [MaxPooling(s) for s in max_pool]
        self.upsampling = [
            nn.Upsample(scale_factor=s, mode="nearest") for s in max_pool
        ]

        # configure encoder (down-sampling path)
        prev_nf = 1
        encoder_nfs = [prev_nf]
        self.encoder = nn.ModuleList()
        for level in range(self.nb_levels - 1):
            convs = nn.ModuleList()
            for conv in range(nb_conv_per_level):
                nf = enc_nf[level * nb_conv_per_level + conv]
                convs.append(ConvBlock(ndims, prev_nf, nf))
                prev_nf = nf
            self.encoder.append(convs)
            encoder_nfs.append(prev_nf)

        # configure decoder (up-sampling path)
        encoder_nfs = np.flip(encoder_nfs)
        self.decoder = nn.ModuleList()
        for level in range(self.nb_levels - 1):
            convs = nn.ModuleList()
            for conv in range(nb_conv_per_level):
                nf = dec_nf[level * nb_conv_per_level + conv]
                convs.append(ConvBlock(ndims, prev_nf, nf))
                prev_nf = nf
            self.decoder.append(convs)
            if level < (self.nb_levels - 1):
                prev_nf += encoder_nfs[level]

        # now we take care of any remaining convolutions
        self.remaining = nn.ModuleList()
        for num, nf in enumerate(final_convs):
            self.remaining.append(ConvBlock(ndims, prev_nf, nf))
            prev_nf = nf

        # final convolutions
        if return_mask:
            self.remaining.append(ConvBlock(ndims, prev_nf, 2, activation=None))
            self.remaining.append(nn.Softmax(dim=1))
        else:
            self.remaining.append(ConvBlock(ndims, prev_nf, 1, activation=None))

    def forward(self, x):

        # encoder forward pass
        x_history = [x]
        for level, convs in enumerate(self.encoder):
            for conv in convs:
                x = conv(x)
            x_history.append(x)
            x = self.pooling[level](x)

        # decoder forward pass with upsampling and concatenation
        for level, convs in enumerate(self.decoder):
            for conv in convs:
                x = conv(x)
            if level < (self.nb_levels - 1):
                x = self.upsampling[level](x)
                x = torch.cat([x, x_history.pop()], dim=1)

        # remaining convs at full resolution
        for conv in self.remaining:
            x = conv(x)

        return x


class ConvBlock(nn.Module):
    """
    Specific convolutional block followed by leakyrelu for unet.
    """

    def __init__(self, ndims, in_channels, out_channels, stride=1, activation="leaky"):
        super().__init__()

        Conv = getattr(nn, "Conv%dd" % ndims)
        self.conv = Conv(in_channels, out_channels, 3, stride, 1)
        if activation == "leaky":
            self.activation = nn.LeakyReLU(0.2)
        elif activation is None:
            self.activation = None
        else:
            raise ValueError(f"Unknown activation: {activation}")

    def forward(self, x):
        out = self.conv(x)
        if self.activation is not None:
            out = self.activation(out)
        return out



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Definition of the command line interface's (CLI) entry point."""


def main():
    """Entry point for MRIQC's CLI."""
    import gc
    import os
    import sys
    from tempfile import mktemp
    import atexit
    from mriqc import config, messages
    from mriqc.cli.parser import parse_args

    atexit.register(config.restore_env)

    # Run parser
    parse_args()

    if config.execution.pdb:
        from mriqc.utils.debug import setup_exceptionhook

        setup_exceptionhook()
        config.nipype.plugin = "Linear"

    # CRITICAL Save the config to a file. This is necessary because the execution graph
    # is built as a separate process to keep the memory footprint low. The most
    # straightforward way to communicate with the child process is via the filesystem.
    # The config file name needs to be unique, otherwise multiple mriqc instances
    # will create write conflicts.
    config_file = mktemp(
        dir=config.execution.work_dir, prefix=".mriqc.", suffix=".toml"
    )
    config.to_filename(config_file)

    # Set up participant level
    if "participant" in config.workflow.analysis_level:
        _pool = None
        if config.nipype.plugin in ("MultiProc", "LegacyMultiProc"):
            from contextlib import suppress
            import multiprocessing as mp
            import multiprocessing.forkserver
            from concurrent.futures import ProcessPoolExecutor

            os.environ["OMP_NUM_THREADS"] = "1"

            with suppress(RuntimeError):
                mp.set_start_method("fork")
            gc.collect()

            _pool = ProcessPoolExecutor(
                max_workers=config.nipype.nprocs,
                initializer=config._process_initializer,
                initargs=(config.execution.cwd, config.nipype.omp_nthreads),
            )

        _resmon = None
        if config.execution.resource_monitor:
            from mriqc.instrumentation.resources import ResourceRecorder

            _resmon = ResourceRecorder(
                pid=os.getpid(),
                log_file=mktemp(
                    dir=config.execution.work_dir, prefix=".resources.", suffix=".tsv"
                ),
            )
            _resmon.start()

        # CRITICAL Call build_workflow(config_file, retval) in a subprocess.
        # Because Python on Linux does not ever free virtual memory (VM), running the
        # workflow construction jailed within a process preempts excessive VM buildup.
        from multiprocessing import Manager, Process

        with Manager() as mgr:
            from .workflow import build_workflow

            retval = mgr.dict()
            p = Process(target=build_workflow, args=(str(config_file), retval))
            p.start()
            p.join()

            mriqc_wf = retval.get("workflow", None)
            retcode = p.exitcode or retval.get("return_code", 0)

        # CRITICAL Load the config from the file. This is necessary because the ``build_workflow``
        # function executed constrained in a process may change the config (and thus the global
        # state of MRIQC).
        config.load(config_file)

        retcode = retcode or (mriqc_wf is None) * os.EX_SOFTWARE
        if retcode != 0:
            sys.exit(retcode)

        # Initalize nipype config
        config.nipype.init()
        # Make sure loggers are started
        config.loggers.init()

        if _resmon:
            config.loggers.cli.info(
                f"Started resource recording at {_resmon._logfile}."
            )

        # Resource management options
        if config.nipype.plugin in ("MultiProc", "LegacyMultiProc") and (
            1 < config.nipype.nprocs < config.nipype.omp_nthreads
        ):
            config.loggers.cli.warning(
                "Per-process threads (--omp-nthreads=%d) exceed total "
                "threads (--nthreads/--n_cpus=%d)",
                config.nipype.omp_nthreads,
                config.nipype.nprocs,
            )

        if mriqc_wf is None:
            sys.exit(os.EX_SOFTWARE)

        if mriqc_wf and config.execution.write_graph:
            mriqc_wf.write_graph(graph2use="colored", format="svg", simple_form=True)

        if not config.execution.dry_run:
            # Warn about submitting measures BEFORE
            if not config.execution.no_sub:
                config.loggers.cli.warning(config.DSA_MESSAGE)

            # Clean up master process before running workflow, which may create forks
            gc.collect()
            # run MRIQC
            _plugin = config.nipype.get_plugin()
            if _pool:
                from mriqc.engine.plugin import MultiProcPlugin

                _plugin = {
                    "plugin": MultiProcPlugin(
                        pool=_pool, plugin_args=config.nipype.plugin_args
                    ),
                }
            mriqc_wf.run(**_plugin)

            # Warn about submitting measures AFTER
            if not config.execution.no_sub:
                config.loggers.cli.warning(config.DSA_MESSAGE)
        config.loggers.cli.log(25, messages.PARTICIPANT_FINISHED)

        if _resmon is not None:
            from mriqc.instrumentation.viz import plot
            _resmon.stop()
            plot(
                _resmon._logfile,
                param="mem_rss_mb",
                out_file=str(_resmon._logfile).replace(".tsv", ".rss.png"),
            )
            plot(
                _resmon._logfile,
                param="mem_vsm_mb",
                out_file=str(_resmon._logfile).replace(".tsv", ".vsm.png"),
            )

    # Set up group level
    if "group" in config.workflow.analysis_level:
        from ..reports import group_html
        from ..utils.bids import DEFAULT_TYPES
        from ..utils.misc import generate_tsv  # , generate_pred

        config.loggers.cli.info(messages.GROUP_START)

        # Generate reports
        mod_group_reports = []
        for mod in config.execution.modalities or DEFAULT_TYPES:
            output_dir = config.execution.output_dir
            dataframe, out_tsv = generate_tsv(output_dir, mod)
            # If there are no iqm.json files, nothing to do.
            if dataframe is None:
                continue

            tsv_message = messages.TSV_GENERATED.format(modality=mod, path=out_tsv)
            config.loggers.cli.info(tsv_message)

            # out_pred = generate_pred(derivatives_dir, settings['output_dir'], mod)
            # if out_pred is not None:
            #     log.info('Predicted QA CSV table for the %s data generated (%s)',
            #                    mod, out_pred)

            out_html = output_dir / f"group_{mod}.html"
            group_html(
                out_tsv,
                mod,
                csv_failed=output_dir / f"group_variant-failed_{mod}.csv",
                out_file=out_html,
            )
            report_message = messages.GROUP_REPORT_GENERATED.format(
                modality=mod, path=out_html
            )
            config.loggers.cli.info(report_message)
            mod_group_reports.append(mod)

        if not mod_group_reports:
            raise Exception(messages.GROUP_NO_DATA)

        config.loggers.cli.info(messages.GROUP_FINISHED)

    from mriqc.utils.bids import write_bidsignore, write_derivative_description

    config.loggers.cli.info(messages.BIDS_META)
    write_derivative_description(config.execution.bids_dir, config.execution.output_dir)
    write_bidsignore(config.execution.output_dir)
    config.loggers.cli.info(messages.RUN_FINISHED)


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
r"""
A Python module to maintain unique, run-wide *MRIQC* settings.

This module implements the memory structures to keep a consistent, singleton config.
Settings are passed across processes via filesystem, and a copy of the settings for
each run and subject is left under
``<output_dir>/sub-<participant_id>/log/<run_unique_id>/mriqc.toml``.
Settings are stored using :abbr:`ToML (Tom's Markup Language)`.
The module has a :py:func:`~mriqc.config.to_filename` function to allow writting out
the settings to hard disk in *ToML* format, which looks like:

.. literalinclude:: ../mriqc/data/config-example.toml
   :language: toml
   :name: mriqc.toml
   :caption: **Example file representation of MRIQC settings**.

This config file is used to pass the settings across processes,
using the :py:func:`~mriqc.config.load` function.

Configuration sections
----------------------
.. autoclass:: environment
   :members:
.. autoclass:: execution
   :members:
.. autoclass:: workflow
   :members:
.. autoclass:: nipype
   :members:

Usage
-----
A config file is used to pass settings and collect information as the execution
graph is built across processes.

.. code-block:: Python

    from mriqc import config
    config_file = mktemp(dir=config.execution.work_dir, prefix='.mriqc.', suffix='.toml')
    config.to_filename(config_file)
    # Call build_workflow(config_file, retval) in a subprocess
    with Manager() as mgr:
        from .workflow import build_workflow
        retval = mgr.dict()
        p = Process(target=build_workflow, args=(str(config_file), retval))
        p.start()
        p.join()
    config.load(config_file)
    # Access configs from any code section as:
    value = config.section.setting

Logging
-------
.. autoclass:: loggers
   :members:

Other responsibilities
----------------------
The :py:mod:`config` is responsible for other conveniency actions.

  * Switching Python's :obj:`multiprocessing` to *forkserver* mode.
  * Set up a filter for warnings as early as possible.
  * Automated I/O magic operations. Some conversions need to happen in the
    store/load processes (e.g., from/to :obj:`~pathlib.Path` \<-\> :obj:`str`,
    :py:class:`~bids.layout.BIDSLayout`, etc.)

"""
import os
import sys
from pathlib import Path
from time import strftime
from uuid import uuid4

try:
    # This option is only available with Python 3.8
    from importlib.metadata import version as get_version
except ImportError:
    from importlib_metadata import version as get_version

# Ignore annoying warnings
from mriqc._warnings import logging

__version__ = get_version("mriqc")
_pre_exec_env = dict(os.environ)

# Reduce numpy's vms by limiting OMP_NUM_THREADS
_default_omp_threads = int(os.getenv("OMP_NUM_THREADS", os.cpu_count()))

# Disable NiPype etelemetry always
_disable_et = bool(
    os.getenv("NO_ET") is not None or os.getenv("NIPYPE_NO_ET") is not None
)
os.environ["NIPYPE_NO_ET"] = "1"
os.environ["NO_ET"] = "1"

if not hasattr(sys, "_is_pytest_session"):
    sys._is_pytest_session = False  # Trick to avoid sklearn's FutureWarnings
# Disable all warnings in main and children processes only on production versions
if not any(
    (
        "+" in __version__,
        __version__.endswith(".dirty"),
        os.getenv("MRIQC_DEV", "0").lower() in ("1", "on", "true", "y", "yes"),
    )
):
    os.environ["PYTHONWARNINGS"] = "ignore"

logging.addLevelName(25, "IMPORTANT")  # Add a new level between INFO and WARNING
logging.addLevelName(15, "VERBOSE")  # Add a new level between INFO and DEBUG

DEFAULT_MEMORY_MIN_GB = 0.01
DSA_MESSAGE = """\
IMPORTANT: Anonymized quality metrics (IQMs) will be submitted to MRIQC's metrics \
repository. \
Submission of IQMs can be disabled using the ``--no-sub`` argument. \
Please visit https://mriqc.readthedocs.io/en/latest/dsa.html to revise MRIQC's \
Data Sharing Agreement."""

_exec_env = os.name
_docker_ver = None
# special variable set in the container
if os.getenv("IS_DOCKER_8395080871"):
    _exec_env = "singularity"
    _cgroup = Path("/proc/1/cgroup")
    if _cgroup.exists() and "docker" in _cgroup.read_text():
        _docker_ver = os.getenv("DOCKER_VERSION_8395080871")
        _exec_env = "docker"
    del _cgroup

_templateflow_home = Path(
    os.getenv(
        "TEMPLATEFLOW_HOME",
        os.path.join(os.getenv("HOME"), ".cache", "templateflow"),
    )
)

try:
    from psutil import virtual_memory

    _free_mem_at_start = round(virtual_memory().free / 1024**3, 1)
except Exception:
    _free_mem_at_start = None

_oc_limit = "n/a"
_oc_policy = "n/a"
try:
    # Memory policy may have a large effect on types of errors experienced
    _proc_oc_path = Path("/proc/sys/vm/overcommit_memory")
    if _proc_oc_path.exists():
        _oc_policy = {"0": "heuristic", "1": "always", "2": "never"}.get(
            _proc_oc_path.read_text().strip(), "unknown"
        )
        if _oc_policy != "never":
            _proc_oc_kbytes = Path("/proc/sys/vm/overcommit_kbytes")
            if _proc_oc_kbytes.exists():
                _oc_limit = _proc_oc_kbytes.read_text().strip()
            if (
                _oc_limit in ("0", "n/a")
                and Path("/proc/sys/vm/overcommit_ratio").exists()
            ):
                _oc_limit = "{}%".format(
                    Path("/proc/sys/vm/overcommit_ratio").read_text().strip()
                )
except Exception:
    pass

_memory_gb = None
try:
    if "linux" in sys.platform:
        with open("/proc/meminfo", "r") as f_in:
            _meminfo_lines = f_in.readlines()
            _mem_total_line = [line for line in _meminfo_lines if "MemTotal" in line][0]
            _mem_total = float(_mem_total_line.split()[1])
            _memory_gb = _mem_total / (1024.0**2)
    elif "darwin" in sys.platform:
        _mem_str = os.popen("sysctl hw.memsize").read().strip().split(" ")[-1]
        _memory_gb = float(_mem_str) / (1024.0**3)
except Exception:
    pass


class _Config:
    """An abstract class forbidding instantiation."""

    _paths = tuple()

    def __init__(self):
        """Avert instantiation."""
        raise RuntimeError("Configuration type is not instantiable.")

    @classmethod
    def load(cls, settings, init=True):
        """Store settings from a dictionary."""
        for k, v in settings.items():
            if v is None:
                continue
            if k in cls._paths:
                setattr(cls, k, Path(v).absolute())
                continue
            if hasattr(cls, k):
                setattr(cls, k, v)

        if init:
            try:
                cls.init()
            except AttributeError:
                pass

    @classmethod
    def get(cls):
        """Return defined settings."""
        out = {}
        for k, v in cls.__dict__.items():
            if k.startswith("_") or v is None:
                continue
            if callable(getattr(cls, k)):
                continue
            if k in cls._paths:
                v = str(v)
            out[k] = v
        return out


class environment(_Config):
    """
    Read-only options regarding the platform and environment.

    Crawls runtime descriptive settings (e.g., default FreeSurfer license,
    execution environment, nipype and *MRIQC* versions, etc.).
    The ``environment`` section is not loaded in from file,
    only written out when settings are exported.
    This config section is useful when reporting issues,
    and these variables are tracked whenever the user does not
    opt-out using the ``--notrack`` argument.

    """

    cpu_count = os.cpu_count()
    """Number of available CPUs."""
    exec_docker_version = _docker_ver
    """Version of Docker Engine."""
    exec_env = _exec_env
    """A string representing the execution platform."""
    free_mem = _free_mem_at_start
    """Free memory at start."""
    overcommit_policy = _oc_policy
    """Linux's kernel virtual memory overcommit policy."""
    overcommit_limit = _oc_limit
    """Linux's kernel virtual memory overcommit limits."""
    nipype_version = get_version("nipype")
    """Nipype's current version."""
    templateflow_version = get_version("templateflow")
    """The TemplateFlow client version installed."""
    total_memory = _memory_gb
    """Total memory available, in GB."""
    version = __version__
    """*MRIQC*'s version."""
    _pre_mriqc = _pre_exec_env
    """Environment variables before MRIQC's execution."""


class nipype(_Config):
    """Nipype settings."""

    crashfile_format = "txt"
    """The file format for crashfiles, either text or pickle."""
    get_linked_libs = False
    """Run NiPype's tool to enlist linked libraries for every interface."""
    local_hash_check = True
    """Check if interface is cached locally before executing."""
    memory_gb = None
    """Estimation in GB of the RAM this workflow can allocate at any given time."""
    nprocs = os.cpu_count()
    """Number of processes (compute tasks) that can be run in parallel (multiprocessing only)."""
    omp_nthreads = _default_omp_threads
    """Number of CPUs a single process can access for multithreaded execution."""
    plugin = "MultiProc"
    """NiPype's execution plugin."""
    plugin_args = {
        "maxtasksperchild": 1,
        "raise_insufficient": False,
    }
    """Settings for NiPype's execution plugin."""
    remove_node_directories = False
    """Remove directories whose outputs have already been used up."""
    resource_monitor = False
    """Enable resource monitor."""
    stop_on_first_crash = True
    """Whether the workflow should stop or continue after the first error."""

    @classmethod
    def get_plugin(cls):
        """Format a dictionary for Nipype consumption."""
        out = {
            "plugin": cls.plugin,
            "plugin_args": cls.plugin_args,
        }
        if cls.plugin in ("MultiProc", "LegacyMultiProc"):
            out["plugin_args"]["n_procs"] = int(cls.nprocs)
            if cls.memory_gb:
                out["plugin_args"]["memory_gb"] = float(cls.memory_gb)
        return out

    @classmethod
    def init(cls):
        """Set NiPype configurations."""
        from nipype import config as ncfg

        # Nipype config (logs and execution)
        ncfg.update_config(
            {
                "execution": {
                    "crashdump_dir": str(execution.log_dir),
                    "crashfile_format": cls.crashfile_format,
                    "get_linked_libs": cls.get_linked_libs,
                    "stop_on_first_crash": cls.stop_on_first_crash,
                }
            }
        )


class execution(_Config):
    """Configure run-level settings."""

    ants_float = False
    """Use float number precision for ANTs computations."""
    bids_dir = None
    """An existing path to the dataset, which must be BIDS-compliant."""
    bids_database_dir = None
    """Path to the directory containing SQLite database indices for the input BIDS dataset."""
    bids_description_hash = None
    """Checksum (SHA256) of the ``dataset_description.json`` of the BIDS dataset."""
    cwd = os.getcwd()
    """Current working directory."""
    debug = False
    """Run in sloppy mode (meaning, suboptimal parameters that minimize run-time)."""
    dry_run = False
    """Just test, do not run."""
    dsname = "<unset>"
    """A dataset name used when generating files from the rating widget."""
    echo_id = None
    """Select a particular echo for multi-echo EPI datasets."""
    float32 = True
    """Cast the input data to float32 if it's represented whith higher precision."""
    layout = None
    """A :py:class:`~bids.layout.BIDSLayout` object, see :py:func:`init`."""
    log_dir = None
    """The path to a directory that contains execution logs."""
    log_level = 25
    """Output verbosity."""
    modalities = None
    """Filter input dataset by MRI type."""
    no_sub = False
    """Turn off submission of anonymized quality metrics to Web API."""
    output_dir = None
    """Folder where derivatives will be stored."""
    participant_label = None
    """List of participant identifiers that are to be preprocessed."""
    pdb = False
    """Drop into PDB when exceptions are encountered."""
    reports_only = False
    """Only build the reports, based on the reportlets found in a cached working directory."""
    resource_monitor = False
    """Enable resource monitor."""
    run_id = None
    """Filter input dataset by run identifier."""
    run_uuid = "%s_%s" % (strftime("%Y%m%d-%H%M%S"), uuid4())
    """Unique identifier of this particular run."""
    session_id = None
    """Filter input dataset by session identifier."""
    task_id = None
    """Select a particular task from all available in the dataset."""
    templateflow_home = _templateflow_home
    """The root folder of the TemplateFlow client."""
    upload_strict = False
    """Workflow will crash if upload is not successful."""
    verbose_reports = False
    """Generate extended reports."""
    webapi_url = "https://mriqc.nimh.nih.gov/api/v1"
    """IP address where the MRIQC WebAPI is listening."""
    webapi_port = None
    """port where the MRIQC WebAPI is listening."""
    work_dir = Path("work").absolute()
    """Path to a working directory where intermediate results will be available."""
    write_graph = False
    """Write out the computational graph corresponding to the planned preprocessing."""

    _layout = None

    _paths = (
        "anat_derivatives",
        "bids_dir",
        "bids_database_dir",
        "fs_license_file",
        "fs_subjects_dir",
        "layout",
        "log_dir",
        "output_dir",
        "templateflow_home",
        "work_dir",
    )

    @classmethod
    def init(cls):
        """Create a new BIDS Layout accessible with :attr:`~execution.layout`."""
        if cls._layout is None:
            import re
            from bids.layout.index import BIDSLayoutIndexer
            from bids.layout import BIDSLayout

            _db_path = cls.bids_database_dir or (
                cls.work_dir / cls.run_uuid / "bids_db"
            )
            _db_path.mkdir(exist_ok=True, parents=True)

            # Recommended after PyBIDS 12.1
            _indexer = BIDSLayoutIndexer(
                validate=False,
                ignore=(
                    "code",
                    "stimuli",
                    "sourcedata",
                    "models",
                    "derivatives",
                    "scripts",
                    re.compile(r"^\."),
                    # Exclude modalities and contrasts ignored by MRIQC (doesn't know how to QC)
                    re.compile(
                        r"sub-[a-zA-Z0-9]+(/ses-[a-zA-Z0-9]+)?/(dwi|fmap|perf)/"
                    ),
                    re.compile(
                        r"sub-[a-zA-Z0-9]+(/ses-[a-zA-Z0-9]+)?/anat/.*_"
                        r"(PDw|T2starw|FLAIR|inplaneT1|inplaneT2|PDT2|angio|T2star"
                        r"|FLASH|PD|T1map|T2map|T2starmap|R1map|R2map|R2starmap|PDmap"
                        r"|MTRmap|MTsat|UNIT1|T1rho|MWFmap|MTVmap|PDT2map|Chimap"
                        r"|S0map|M0map|defacemask|MESE|MEGRE|VFA|IRT1|MP2RAGE|MPM|MTS|MTR)\."
                    ),
                    re.compile(
                        r"sub-[a-zA-Z0-9]+(/ses-[a-zA-Z0-9]+)?/func/.*"
                        r"_(cbv|sbref|phase|events|physio|stim)\."
                    ),
                ),
            )
            cls._layout = BIDSLayout(
                str(cls.bids_dir),
                database_path=_db_path,
                reset_database=cls.bids_database_dir is None,
                indexer=_indexer,
            )
            cls.bids_database_dir = _db_path

        cls.layout = cls._layout


# These variables are not necessary anymore
del _exec_env
del _templateflow_home
del _free_mem_at_start
del _oc_limit
del _oc_policy


class workflow(_Config):
    """Configure the particular execution graph of this workflow."""

    analysis_level = ["participant"]
    """Level of analysis."""
    biggest_file_gb = 1
    """Size of largest file in GB."""
    deoblique = False
    """Deoblique the functional scans during head motion correction preprocessing."""
    despike = False
    """Despike the functional scans during head motion correction preprocessing."""
    fd_thres = 0.2
    """Threshold on Framewise Displacement estimates to detect outliers."""
    fd_radius = 50
    """Radius in mm. of the sphere for the FD calculation."""
    fft_spikes_detector = False
    """Turn on FFT based spike detector (slow)."""
    headmask = "BET"
    """Use FSL BET in :py:func:`~mriqc.workflows.anatomical.headmsk_wf`."""
    ica = False
    """Run ICA on the raw data and include the components in the individual reports."""
    inputs = None
    """List of files to be processed with MRIQC."""
    species = "human"
    """Subject species to choose most appropriate template"""
    template_id = "MNI152NLin2009cAsym"
    """TemplateFlow ID of template used for the anatomical processing."""


class loggers:
    """Keep loggers easily accessible (see :py:func:`init`)."""

    _fmt = "%(asctime)s,%(msecs)d %(name)-2s " "%(levelname)-2s:\n\t %(message)s"
    _datefmt = "%y%m%d-%H:%M:%S"
    _init = False

    default = logging.getLogger()
    """The root logger."""
    cli = logging.getLogger("cli")
    """Command-line interface logging."""
    workflow = None
    """NiPype's workflow logger."""
    interface = None
    """NiPype's interface logger."""
    utils = None
    """NiPype's utils logger."""

    @classmethod
    def init(cls):
        """
        Set the log level, initialize all loggers into :py:class:`loggers`.

            * Add new logger levels (25: IMPORTANT, and 15: VERBOSE).
            * Add a new sub-logger (``cli``).
            * Logger configuration.

        """
        if not cls._init:
            from nipype import logging as nlogging
            from nipype import config as ncfg

            cls.workflow = nlogging.getLogger("nipype.workflow")
            cls.interface = nlogging.getLogger("nipype.interface")
            cls.utils = nlogging.getLogger("nipype.utils")

            if not len(cls.cli.handlers):
                _handler = logging.StreamHandler(stream=sys.stdout)
                _handler.setFormatter(
                    logging.Formatter(fmt=cls._fmt, datefmt=cls._datefmt)
                )
                cls.cli.addHandler(_handler)

            ncfg.update_config(
                {
                    "logging": {
                        "log_directory": str(execution.log_dir),
                        "log_to_file": True,
                    },
                }
            )
            cls._init = True

        cls.default.setLevel(execution.log_level)
        cls.cli.setLevel(execution.log_level)
        cls.interface.setLevel(execution.log_level)
        cls.workflow.setLevel(execution.log_level)
        cls.utils.setLevel(execution.log_level)

    @classmethod
    def getLogger(cls, name):
        """Create a new logger."""
        retval = getattr(cls, name)
        if retval is None:
            setattr(cls, name, logging.getLogger(name))
            _handler = logging.StreamHandler(stream=sys.stdout)
            _handler.setFormatter(logging.Formatter(fmt=cls._fmt, datefmt=cls._datefmt))
            retval.addHandler(_handler)
            retval.setLevel(execution.log_level)
        return retval


def from_dict(settings):
    """Read settings from a flat dictionary."""
    execution.load(settings)
    workflow.load(settings)
    nipype.load(settings, init=False)


def load(filename):
    """Load settings from file."""
    from toml import loads

    filename = Path(filename)
    settings = loads(filename.read_text())
    for sectionname, configs in settings.items():
        if sectionname != "environment":
            section = getattr(sys.modules[__name__], sectionname)
            section.load(configs)


def get(flat=False):
    """Get config as a dict."""
    settings = {
        "environment": environment.get(),
        "execution": execution.get(),
        "workflow": workflow.get(),
        "nipype": nipype.get(),
    }
    if not flat:
        return settings

    return {
        ".".join((section, k)): v
        for section, configs in settings.items()
        for k, v in configs.items()
    }


def dumps():
    """Format config into toml."""
    from toml import dumps

    return dumps(get())


def to_filename(filename):
    """Write settings to file."""
    filename = Path(filename)
    filename.parent.mkdir(exist_ok=True, parents=True)
    filename.write_text(dumps())


def _process_initializer(cwd, omp_nthreads):
    """Initialize the environment of the child process."""
    os.chdir(cwd)
    os.environ["NIPYPE_NO_ET"] = "1"
    os.environ["OMP_NUM_THREADS"] = f"{omp_nthreads}"


def restore_env():
    """Restore the original environment."""

    for k in os.environ.keys():
        del os.environ[k]

    for k, v in environment._pre_mriqc.items():
        os.environ[k] = v



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Utilities and mocks for testing and documentation building."""
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkdtemp

from pkg_resources import resource_filename as pkgrf
from toml import loads


@contextmanager
def mock_config():
    """Create a mock config for documentation and testing purposes."""
    from . import config

    filename = Path(pkgrf("mriqc", "data/config-example.toml"))
    settings = loads(filename.read_text())
    for sectionname, configs in settings.items():
        if sectionname != "environment":
            section = getattr(config, sectionname)
            section.load(configs, init=False)
    config.nipype.init()
    config.loggers.init()

    config.execution.work_dir = Path(mkdtemp())
    config.execution.bids_dir = Path(pkgrf("mriqc", "data/tests/ds000005")).absolute()
    config.execution.init()

    yield



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""ABIDE2BIDS download tool."""
import errno
import json
import os
import os.path as op
import shutil
import subprocess as sp
import tempfile
from argparse import ArgumentParser, RawTextHelpFormatter
from multiprocessing import Pool
from typing import Tuple
from xml.etree import ElementTree as et

import numpy as np
from mriqc.bin import messages


def main():
    """Entry point."""
    parser = ArgumentParser(
        description="ABIDE2BIDS downloader.",
        formatter_class=RawTextHelpFormatter,
    )
    g_input = parser.add_argument_group("Inputs")
    g_input.add_argument("-i", "--input-abide-catalog", action="store", required=True)
    g_input.add_argument(
        "-n", "--dataset-name", action="store", default="ABIDE Dataset"
    )
    g_input.add_argument(
        "-u", "--nitrc-user", action="store", default=os.getenv("NITRC_USER")
    )
    g_input.add_argument(
        "-p",
        "--nitrc-password",
        action="store",
        default=os.getenv("NITRC_PASSWORD"),
    )

    g_outputs = parser.add_argument_group("Outputs")
    g_outputs.add_argument("-o", "--output-dir", action="store", default="ABIDE-BIDS")

    opts = parser.parse_args()

    if opts.nitrc_user is None or opts.nitrc_password is None:
        raise RuntimeError("NITRC user and password are required")

    dataset_desc = {
        "BIDSVersion": "1.0.0rc3",
        "License": "CC Attribution-NonCommercial-ShareAlike 3.0 Unported",
        "Name": opts.dataset_name,
    }

    out_dir = op.abspath(opts.output_dir)
    try:
        os.makedirs(out_dir)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise exc

    with open(op.join(out_dir, "dataset_description.json"), "w") as dfile:
        json.dump(dataset_desc, dfile)

    catalog = et.parse(opts.input_abide_catalog).getroot()
    urls = [el.get("URI") for el in catalog.iter() if el.get("URI") is not None]

    pool = Pool()
    args_list = [(url, opts.nitrc_user, opts.nitrc_password, out_dir) for url in urls]
    res = pool.map(fetch, args_list)

    tsv_data = np.array([("subject_id", "site_name")] + res)
    np.savetxt(
        op.join(out_dir, "participants.tsv"),
        tsv_data,
        fmt="%s",
        delimiter="\t",
    )


def fetch(args: Tuple[str, str, str, str]) -> Tuple[str, str]:
    """
    Downloads a subject and formats it into BIDS.

    Parameters
    ----------
    args : Tuple[str, str, str, str]
        URL, NITRC user, NITRC password, destination

    Returns
    -------
    Tuple[str, str]
        Subject ID, Site name
    """
    out_dir = None
    if len(args) == 3:
        url, user, password = args
    else:
        url, user, password, out_dir = args

    tmpdir = tempfile.mkdtemp()
    if out_dir is None:
        out_dir = os.getcwd()
    else:
        out_dir = op.abspath(out_dir)

    pkg_id = [u[9:] for u in url.split("/") if u.startswith("NITRC_IR_")][0]
    sub_file = op.join(tmpdir, "%s.zip" % pkg_id)

    cmd = ["curl", "-s", "-u", "%s:%s" % (user, password), "-o", sub_file, url]
    sp.check_call(cmd)
    sp.check_call(["unzip", "-qq", "-d", tmpdir, "-u", sub_file])

    abide_root = op.join(tmpdir, "ABIDE")
    files = []
    for root, path, fname in os.walk(abide_root):
        if fname and (fname[0].endswith("nii") or fname[0].endswith("nii.gz")):
            if path:
                root = op.join(root, path[0])
            files.append(op.join(root, fname[0]))

    index = len(abide_root) + 1
    site_name, sub_str = files[0][index:].split("/")[0].split("_")
    subject_id = "sub-" + sub_str

    for i in files:
        ext = ".nii.gz"
        if i.endswith(".nii"):
            ext = ".nii"
        if "mprage" in i:
            bids_dir = op.join(out_dir, subject_id, "anat")
            try:
                os.makedirs(bids_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise exc
            shutil.copy(i, op.join(bids_dir, subject_id + "_T1w" + ext))

        if "rest" in i:
            bids_dir = op.join(out_dir, subject_id, "func")
            try:
                os.makedirs(bids_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise exc
            shutil.copy(i, op.join(bids_dir, subject_id + "_rest_bold" + ext))

    shutil.rmtree(tmpdir, ignore_errors=True, onerror=_myerror)

    success_message = messages.ABIDE_SUBJECT_FETCHED.format(
        subject_id=subject_id[4:], site_name=site_name
    )
    print(success_message)
    return subject_id[4:], site_name


def _myerror(message: str):
    """
    Print warning in case an exception is raised for temporal files removal.

    Parameters
    ----------
    message : str
        `shutil.rmtree()` error message
    """
    warning = messages.ABIDE_TEMPORAL_WARNING.format(message=message)
    print(warning)


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2022 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""SynthStrip interface."""
import os
from pathlib import Path
from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    File,
    TraitedSpec,
    traits,
    Undefined,
)

_fs_home = os.getenv("FREESURFER_HOME", None)
_default_model_path = Path(_fs_home) / "models" / "synthstrip.1.pt" if _fs_home else Undefined

if _fs_home and not _default_model_path.exists():
    _default_model_path = Undefined


class _SynthStripInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="-i %s",
        desc="Input image to be brain extracted",
    )
    use_gpu = traits.Bool(
        False, usedefault=True, argstr="-g", desc="Use GPU", nohash=True
    )
    model = File(
        str(_default_model_path),
        usedefault=True,
        exists=True,
        argstr="--model %s",
        desc="file containing model's weights",
    )
    border_mm = traits.Int(
        1, usedefault=True, argstr="-b %d", desc="Mask border threshold in mm"
    )
    out_file = File(
        name_source=["in_file"],
        name_template="%s_desc-brain.nii.gz",
        argstr="-o %s",
        desc="store brain-extracted input to file",
    )
    out_mask = File(
        name_source=["in_file"],
        name_template="%s_desc-brain_mask.nii.gz",
        argstr="-m %s",
        desc="store brainmask to file",
    )


class _SynthStripOutputSpec(TraitedSpec):
    out_file = File(desc="brain-extracted image")
    out_mask = File(desc="brain mask")


class SynthStrip(CommandLine):
    _cmd = "synthstrip"
    input_spec = _SynthStripInputSpec
    output_spec = _SynthStripOutputSpec
